# We use double underscore to make it the first test alphabetically.
import pytest
import pyccl as ccl
import numpy as np
from time import time

# Enable caching for this test.
DEFAULT_CACHING_STATUS = ccl.Caching._enabled

NUM = 3  # number of different Cosmologies we will check
# some unusual numbers that have not occurred before
s8_arr = np.linspace(0.753141592, 0.953141592, NUM)
# a modest speed increase - we are modest in the test to accommodate for slow
# runs; normally this is is expected to be another order of magnitude faster
SPEEDUP = 50


def get_cosmo(sigma8):
    return ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, h=0.67, n_s=0.96,
                         sigma8=sigma8)


@ccl.cache(maxsize=3)
def cosmo_create_and_compute_linpow(sigma8):
    cosmo = get_cosmo(sigma8)
    cosmo.compute_linear_power()
    return cosmo


def timeit_(sigma8):
    t0 = time()
    cosmo_create_and_compute_linpow(sigma8)
    return time() - t0


def test_caching_switches():
    """Test that the Caching switches work as intended."""
    assert ccl.Caching._maxsize == ccl.Caching._default_maxsize
    ccl.Caching.maxsize = 128
    assert ccl.Caching._maxsize == 128
    ccl.Caching.disable()
    assert not ccl.Caching._enabled
    ccl.Caching.enable()
    assert ccl.Caching._enabled
    ccl.Caching.disable()


def test_times():
    """Verify that caching is happening.
    Return time for querying the Boltzmann code goes from O(1s) to O(5ms).
    """
    # If we disable caching, t1 and t2 will be of the same order of magnitude.
    # No need to run through the entire s8_arr.
    ccl.Caching.disable()
    t1 = np.array([timeit_(s8) for s8 in s8_arr[:1]])
    t2 = np.array([timeit_(s8) for s8 in s8_arr[:1]])
    assert np.abs(np.log10(t2/t1)) < 1.0
    # But if caching is enabled, the second call will be much faster.
    ccl.Caching.enable()
    t1 = np.array([timeit_(s8) for s8 in s8_arr])
    t2 = np.array([timeit_(s8) for s8 in s8_arr])
    assert np.all(t1/t2 > SPEEDUP)
    ccl.Caching.disable()


def test_caching_fifo():
    """Test First-In-First-Out retention policy."""
    ccl.Caching.enable()
    # To save time, we test caching by limiting the maximum cache size
    # from 64 (default) to 3. We cache Comologies with different sigma8.
    # By now, the caching repo will be full.
    ccl.Caching.maxsize = NUM
    func = cosmo_create_and_compute_linpow
    assert len(func.cache_info._caches) >= ccl.Caching.maxsize

    ccl.Caching.policy = "fifo"

    t1 = timeit_(sigma8=s8_arr[0])  # this is the oldest cached pk
    # create new and discard oldest
    cosmo_create_and_compute_linpow(0.42)
    t2 = timeit_(sigma8=s8_arr[0])  # cached again
    assert t2/t1 > SPEEDUP
    ccl.Caching.disable()


def test_caching_lru():
    """Test Least-Recently-Used retention policy."""
    # By now the stored Cosmologies are { s8_arr[2], 0.42, s8_arr[0]] }
    # from oldest to newest. Here, we show that we can retain s8_arr[2]
    # simply by using it and moving it to the end of the stack.
    ccl.Caching.enable()
    ccl.Caching.policy = "lru"

    t1 = timeit_(sigma8=s8_arr[2])  # moves to the end of the stack
    # create new and discard the least recently used
    cosmo_create_and_compute_linpow(0.43)
    t2 = timeit_(sigma8=s8_arr[2])  # retrieved
    assert np.abs(np.log10(t2/t1)) < 1.0
    ccl.Caching.disable()


def test_caching_lfu():
    """Test Least-Frequently-Used retention policy."""
    # Now, the stored Cosmologies are { s8_arr[0], 0.43, s8_arr[2] }
    # from oldest to newest. Here, we call each a different number of times
    # and we check that the one used the least (0.43) is discarded.
    ccl.Caching.enable()
    ccl.Caching.policy = "lfu"

    t1 = timeit_(sigma8=0.43)  # increments counter by 1
    _ = [timeit_(sigma8=s8_arr[0]) for _ in range(5)]
    _ = [timeit_(sigma8=s8_arr[2]) for _ in range(3)]
    # create new and discard the least frequently used
    cosmo_create_and_compute_linpow(0.44)
    t2 = timeit_(sigma8=0.43)  # cached again
    assert t2/t1 > SPEEDUP
    ccl.Caching.disable()


def test_cache_info():
    """Test that the CacheInfo repr gives us the expected information."""
    info = cosmo_create_and_compute_linpow.cache_info
    for text in ["maxsize", "policy", "hits", "misses", "current_size"]:
        assert text in repr(info)

    obj = list(info._caches.values())[0]
    assert "counter" in repr(obj)


def test_caching_reset():
    """Test the reset switches."""
    ccl.Caching.reset()
    assert ccl.Caching.maxsize == ccl.Caching._default_maxsize
    assert ccl.Caching.policy == ccl.Caching._default_policy
    ccl.Caching.clear_cache()
    func = cosmo_create_and_compute_linpow
    assert len(func.cache_info._caches) == 0


def test_caching_policy_raises():
    """Test that if the set policy is not correct, it raises an exception."""
    with pytest.raises(ValueError):
        @ccl.Caching.cache(maxsize=-1)
        def func1():
            return

    with pytest.raises(ValueError):
        @ccl.Caching.cache(policy="my_policy")
        def func2():
            return

    with pytest.raises(ValueError):
        ccl.Caching.maxsize = -1

    with pytest.raises(ValueError):
        ccl.Caching.policy = "my_policy"


def test_caching_parentheses():
    """Verify that ``cache`` can be used with or without parentheses."""
    func1, func2 = [lambda: None for _ in range(2)]
    ccl.cache(func1)
    ccl.cache(maxsize=10, policy="lru")(func2)
    assert all([hasattr(func, "cache_info") for func in [func1, func2]])


# Revert to defaults.
ccl.Caching._enabled = DEFAULT_CACHING_STATUS



import numpy as np
import pyccl as ccl
import pytest


def get_prediction(ells, chi_i, chi_f, alpha, beta, gamma,
                   der_bessel, der_angles):
    exponent = 2 * gamma - alpha - 2 * beta - 1
    fl = np.ones_like(ells)
    if der_bessel == -1:
        fl *= 1. / (ells + 0.5)**4
    if der_angles == 2:
        fl *= (ells + 2.) * (ells + 1.) * ells * (ells - 1)
    elif der_angles == 1:
        fl *= ((ells + 1.) * ells)**2
    lfac = fl * (ells + 0.5)**(alpha + 2 * beta)
    norm = (chi_f**exponent-chi_i**exponent) / exponent
    return lfac * norm


@pytest.fixture(scope='module')
def set_up():
    ccl.physical_constants.unfreeze()
    ccl.physical_constants.T_CMB = 2.7
    ccl.gsl_params.INTEGRATION_LIMBER_EPSREL = 1E-4
    ccl.gsl_params.INTEGRATION_EPSREL = 1E-4
    cosmo = ccl.Cosmology(Omega_c=0.30, Omega_b=0.00, Omega_g=0, Omega_k=0,
                          h=0.7, sigma8=0.8, n_s=0.96, Neff=0, m_nu=0.0,
                          w0=-1, wa=0, transfer_function='bbks',
                          matter_power_spectrum='linear')

    ccl.physical_constants.reload()
    ccl.gsl_params.reload()
    return cosmo


@pytest.mark.parametrize("alpha,beta,gamma,is_factorizable,"
                         "w_transfer,mismatch,der_bessel,der_angles",
                         # non-factorizable
                         [(-2., -1., -1., False, True, False, 0, 0),
                          (-2., 0., -1., False, True, False, 0, 0),
                          (-2., -1., 0., False, True, False, 0, 0),
                          (-2., 0., 0., False, True, False, 0, 0),
                          (0., 0., 0., False, True, False, 0, 0),
                          # factorizable transfer functions
                          (-2., -1., -1., True, True, False, 0, 0),
                          (-2., 0., -1., True, True, False, 0, 0),
                          (-2., -1., 0., True, True, False, 0, 0),
                          (-2., 0., 0., True, True, False, 0, 0),
                          (0., 0., 0., True, True, False, 0, 0),
                          # Unit transfer functions
                          (-2., 0., 0., False, False, False, 0, 0),
                          (0., 0., 0., False, False, False, 0, 0),
                          # Mismatch in kernel-transfer coverage
                          (-2., -1., 0., True, True, True, 0, 0),
                          (-2., -1., 0., False, True, True, 0, 0),
                          # non-zero der_bessel
                          (-2., -1., -1., False, True, False, -1, 0),
                          # non-zero der_angles
                          (-2., -1., -1., False, True, False, 0, 1),
                          (-2., -1., -1., False, True, False, 0, 2),
                          (-2., -1., -1., False, True, False, -1, 2)])
def test_tracers_analytic(set_up, alpha, beta, gamma,
                          is_factorizable, w_transfer,
                          mismatch, der_bessel, der_angles):
    cosmo = set_up
    zmax = 0.8
    zi = 0.4
    zf = 0.6
    nchi = 1024
    nk = 512

    # x arrays
    chii = ccl.comoving_radial_distance(cosmo, 1. / (1 + zi))
    chif = ccl.comoving_radial_distance(cosmo, 1. / (1 + zf))
    chimax = ccl.comoving_radial_distance(cosmo, 1. / (1 + zmax))
    chiarr = np.linspace(0.1, chimax, nchi)
    if mismatch:
        # Remove elements around the edges
        mask = (chiarr < 0.9 * chif) & (chiarr > 1.1 * chii)
        chiarr_transfer = chiarr[mask]
    else:
        chiarr_transfer = chiarr
    aarr_transfer = ccl.scale_factor_of_chi(cosmo, chiarr_transfer)[::-1]
    aarr = ccl.scale_factor_of_chi(cosmo, chiarr)[::-1]
    lkarr = np.log(10.**np.linspace(-6, 3, nk))

    # Kernel
    wchi = np.ones_like(chiarr)
    wchi[chiarr < chii] = 0
    wchi[chiarr > chif] = 0

    # Transfer
    t = ccl.Tracer()
    if w_transfer:
        ta = (chiarr_transfer**gamma)[::-1]
        tk = np.exp(beta*lkarr)
        if is_factorizable:
            # 1D
            t.add_tracer(cosmo,
                         kernel=(chiarr, wchi),
                         transfer_k=(lkarr, tk),
                         transfer_a=(aarr_transfer, ta),
                         der_bessel=der_bessel,
                         der_angles=der_angles)
        else:
            # 2D
            tka = ta[:, None] * tk[None, :]
            t.add_tracer(cosmo,
                         kernel=(chiarr, wchi),
                         transfer_ka=(aarr_transfer, lkarr, tka),
                         der_bessel=der_bessel,
                         der_angles=der_angles)
    else:
        t.add_tracer(cosmo,
                     kernel=(chiarr, wchi),
                     der_bessel=der_bessel,
                     der_angles=der_angles)

    # Power spectrum
    pkarr = np.ones_like(aarr)[:, None] * (np.exp(alpha * lkarr))[None, :]
    pk2d = ccl.Pk2D(a_arr=aarr, lk_arr=lkarr, pk_arr=pkarr, is_logp=False)

    # C_ells
    larr = np.linspace(2, 3000, 100)
    # 1D
    cl = ccl.angular_cl(cosmo, t, t, larr, p_of_k_a=pk2d)
    # Prediction
    clpred = get_prediction(larr, chii, chif,
                            alpha, beta, gamma,
                            der_bessel, der_angles)

    assert np.all(np.fabs(cl / clpred - 1) < 5E-3)



import os
import numpy as np
import pyccl as ccl


def test_ssc_WL():
    # Compare against Benjamin Joachimi's code. An overview of the methodology
    # is given in appendix E.2 of 2007.01844.
    data_dir = os.path.join(os.path.dirname(__file__), "data/covariances/")

    h = 0.7
    cosmo = ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, h=h, n_s=0.97,
                          sigma8=0.8, m_nu=0.0)

    mass_def = ccl.halos.MassDef200m
    hmf = ccl.halos.MassFuncTinker10(mass_def=mass_def)
    hbf = ccl.halos.HaloBiasTinker10(mass_def=mass_def)
    con = ccl.halos.ConcentrationDuffy08(mass_def=mass_def)
    nfw = ccl.halos.HaloProfileNFW(mass_def=mass_def, concentration=con,
                                   fourier_analytic=True)
    hmc = ccl.halos.HMCalculator(mass_function=hmf, halo_bias=hbf,
                                 mass_def=mass_def)

    n_z = 100

    n_k = 200
    k_min = 1e-4
    k_max = 1e2

    a = np.linspace(1/(1+6), 1, n_z)
    k = np.geomspace(k_min, k_max, n_k)

    tk3D = ccl.halos.halomod_Tk3D_SSC(cosmo=cosmo, hmc=hmc,
                                      prof=nfw, prof2=nfw, prof12_2pt=None,
                                      lk_arr=np.log(k), a_arr=a,
                                      use_log=True)

    z, nofz = np.loadtxt(os.path.join(data_dir, "ssc_WL_nofz.txt"),
                         unpack=True)
    WL_tracer = ccl.WeakLensingTracer(cosmo, dndz=(z, nofz))

    ell = np.loadtxt(os.path.join(data_dir, "ssc_WL_ell.txt"))

    fsky = 0.05

    sigma2_B = ccl.sigma2_B_disc(cosmo, a_arr=a, fsky=fsky)
    cov_ssc = ccl.covariances.angular_cl_cov_SSC(cosmo,
                                                 tracer1=WL_tracer,
                                                 tracer2=WL_tracer,
                                                 ell=ell, t_of_kk_a=tk3D,
                                                 sigma2_B=(a, sigma2_B),
                                                 fsky=None)
    var_ssc_ccl = np.diag(cov_ssc)
    off_diag_1_ccl = np.diag(cov_ssc, k=1)

    cov_ssc_bj = np.loadtxt(os.path.join(data_dir, "ssc_WL_cov_matrix.txt"))

    # At large scales, CCL uses a different convention for the Limber
    # approximation. This factor accounts for this difference
    ccl_limber_shear_fac = np.sqrt((ell-1)*ell*(ell+1)*(ell+2))/(ell+1/2)**2
    cov_ssc_bj_corrected = cov_ssc_bj * np.outer(ccl_limber_shear_fac**2,
                                                 ccl_limber_shear_fac**2)
    var_bj = np.diag(cov_ssc_bj_corrected)
    off_diag_1_bj = np.diag(cov_ssc_bj_corrected, k=1)

    assert np.all(np.fabs(var_ssc_ccl/var_bj - 1) < 3e-2)
    assert np.all(np.fabs(off_diag_1_ccl/off_diag_1_bj - 1) < 3e-2)
    assert np.all(np.fabs(cov_ssc/cov_ssc_bj_corrected - 1) < 3e-2)



import pyccl as ccl
import numpy as np


def test_spline1d():
    cosmo = ccl.CosmologyVanillaLCDM()
    cosmo.compute_distances()

    chi_gsl_spline = cosmo.cosmo.data.chi
    a_arr, chi_arr = ccl.pyutils._get_spline1d_arrays(chi_gsl_spline)
    chi = ccl.comoving_radial_distance(cosmo, a_arr)

    assert np.allclose(chi_arr, chi)


def test_spline2d():
    x = np.linspace(0.1, 1, 10)
    log_y = np.linspace(-3, 1, 20)
    zarr_in = np.outer(x, np.exp(log_y))

    pk2d = ccl.Pk2D(a_arr=x, lk_arr=log_y, pk_arr=zarr_in,
                    is_logp=False)

    pk2d_gsl_spline2d = pk2d.psp.fka
    xarr, yarr, zarr_out_spline = \
        ccl.pyutils._get_spline2d_arrays(pk2d_gsl_spline2d)

    cosmo = ccl.CosmologyVanillaLCDM()
    zarr_out_eval = pk2d(k=np.exp(log_y), a=x[-1], cosmo=cosmo)

    assert np.allclose(x, xarr)
    assert np.allclose(log_y, yarr)
    assert np.allclose(zarr_in, zarr_out_spline)
    assert np.allclose(zarr_in[-1], zarr_out_eval)



import pyccl


def test_cclerror_repr():
    """Check that a CCLError can be built from its repr"""
    e = pyccl.CCLError("blah")
    e2 = eval(repr(e))
    assert str(e2) == str(e)
    assert e2 == e


def test_cclerror_not_equal():
    """Check that a CCLError can be built from its repr"""
    e = pyccl.CCLError("blah")
    e2 = pyccl.CCLError("blahh")
    assert e is not e2
    assert e != e2
    assert hash(e) != hash(e2)


def test_cclwarning_repr():
    """Check that a CCLWarning can be built from its repr"""
    w = pyccl.CCLWarning("blah")
    w2 = eval(repr(w))
    assert str(w2) == str(w)
    assert w2 == w

    v = pyccl.CCLDeprecationWarning("blah")
    v2 = eval(repr(v))
    assert str(v2) == str(v)
    assert v2 == v


def test_cclwarning_not_equal():
    """Check that a CCLWarning can be built from its repr"""
    w = pyccl.CCLWarning("blah")
    w2 = pyccl.CCLWarning("blahh")
    assert w is not w2
    assert w != w2
    assert hash(w) != hash(w2)

    v = pyccl.CCLDeprecationWarning("blah")
    v2 = pyccl.CCLDeprecationWarning("blahh")
    assert v is not v2
    assert v != v2
    assert hash(v) != hash(v2)


def test_ccl_deprecationwarning_switch():
    import warnings

    # check that warnings are enabled by default
    with warnings.catch_warnings(record=True) as w:
        warnings.warn("test", pyccl.CCLDeprecationWarning)
    assert w[0].category == pyccl.CCLDeprecationWarning

    # switch off CCL (future) deprecation warnings
    pyccl.CCLDeprecationWarning.disable()
    with warnings.catch_warnings(record=True) as w:
        warnings.warn("test", pyccl.CCLDeprecationWarning)
    assert len(w) == 0

    # switch back on
    pyccl.CCLDeprecationWarning.enable()



__all__ = ("mass2radius_lagrangian", "convert_concentration", "MassDef",
           "MassDef200m", "MassDef200c", "MassDef500c", "MassDefVir",
           "MassDefFof", "mass_translator",)

from functools import cached_property

import numpy as np

from .. import CCLAutoRepr, CCLNamedClass, lib, check
from . import Concentration, HaloBias, MassFunc


def mass2radius_lagrangian(cosmo, M):
    """ Returns Lagrangian radius for a halo of mass :math:`M`.
    The Lagrangian radius is defined as that enclosing
    the mass of the halo assuming a homogeneous Universe.

    .. math::
        R = \\left(\\frac{3\\,M}{4\\pi\\,\\rho_{M,0}}\\right)^{1/3}

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
        M (:obj:`float` or `array`): halo mass in units of :math:`M_\\odot`.

    Returns:
        (:obj:`float` or `array`): Lagrangian radius in comoving Mpc.
    """
    M_use = np.atleast_1d(M)
    R = (M_use / (4.18879020479 * cosmo.rho_x(1, 'matter')))**(1./3.)
    if np.ndim(M) == 0:
        return R[0]
    return R


def convert_concentration(cosmo, *, c_old, Delta_old, Delta_new):
    """ Computes the concentration parameter for a different mass definition.
    This is done assuming an NFW profile. The output concentration ``c_new`` is
    found by solving the equation:

    .. math::
        f(c_{\\rm old}) \\Delta_{\\rm old} = f(c_{\\rm new}) \\Delta_{\\rm new}

    where

    .. math::
        f(x) = \\frac{x^3}{\\log(1+x) - x/(1+x)}.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
        c_old (:obj:`float` or `array`): concentration to translate from.
        Delta_old (:obj:`float`): Delta parameter associated to the input
            concentration. See description of the :class:`MassDef` class.
        Delta_new (:obj:`float`): Delta parameter associated to the output
            concentration.

    Returns:
        (:obj:`float` or `array`): concentration parameter for the new
        mass definition.
    """
    status = 0
    c_old_use = np.atleast_1d(c_old)
    c_new, status = lib.convert_concentration_vec(cosmo.cosmo,
                                                  Delta_old, c_old_use,
                                                  Delta_new, c_old_use.size,
                                                  status)
    check(status, cosmo=cosmo)

    if np.isscalar(c_old):
        return c_new[0]
    return c_new


class MassDef(CCLAutoRepr, CCLNamedClass):
    """Halo mass definition. Halo masses are defined in terms of an overdensity
    parameter :math:`\\Delta` and an associated density :math:`X` (either the
    matter density or the critical density):

    .. math::
        M_\\Delta = \\frac{4 \\pi}{3} \\Delta\\,\\rho_X\\, R_\\Delta^3

    where :math:`R_\\Delta` is the halo radius. This object also holds methods
    to translate between :math:`R_\\Delta` and :math:`M_\\Delta`, and to
    translate masses between different definitions if a concentration-mass
    relation is provided.

    You may also define halo masses based on a Friends-of-Friends algorithm,
    in which case simply pass ``Delta='fof'`` below.

    Args:
        Delta (:obj:`float`): overdensity parameter. Pass ``'vir'`` if using virial
            overdensity. Pass ``'fof'`` for Friends-of-Friends halos.
        rho_type (:obj:`str`): either 'critical' or 'matter'.
    """ # noqa
    __eq_attrs__ = ("name",)

    def __init__(self, Delta, rho_type):
        # Check it makes sense
        if isinstance(Delta, str):
            if Delta.isdigit():
                Delta = int(Delta)
            elif Delta not in ["fof", "vir"]:
                raise ValueError(f"Unknown Delta type {Delta}.")
        if isinstance(Delta, (int, float)) and Delta < 0:
            raise ValueError("Delta must be a positive number.")
        if rho_type not in ['matter', 'critical']:
            raise ValueError("rho_type must be {'matter', 'critical'}.")

        self.Delta = Delta
        self.rho_type = rho_type

    @cached_property
    def name(self):
        """Give a name to this mass definition."""
        if isinstance(self.Delta, (int, float)):
            return f"{self.Delta}{self.rho_type[0]}"
        return f"{self.Delta}"

    def __repr__(self):
        return f"MassDef(Delta={self.Delta}, rho_type={self.rho_type})"

    def get_Delta_vir(self, cosmo, a):
        """ Computes the virial collapse density contrast with respect
        to the critical density assuming a :math:`\\Lambda` CDM model. We
        use the fitting function from
        `Bryan and Norman 1998 <https://arxiv.org/abs/astro-ph/9710107>`_.
        The virial overdensity is returned for the density type of this
        object's mass definition (e.g. 'critical' or 'matter').

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
            a (:obj:`float`): scale factor

        Returns:
            :obj:`float`: value of the virial overdensity.
        """
        Omz = cosmo.omega_x(a, 'matter')
        x = Omz-1
        # Eq. 6
        Dv = 18*np.pi**2+82*x-39*x**2
        if self.rho_type == 'matter':
            Dv /= Omz
        return Dv

    def get_Delta(self, cosmo, a):
        """ Gets overdensity parameter associated to this mass
        definition.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
            a (:obj:`float`): scale factor

        Returns:
            :obj:`float`: value of the overdensity parameter.
        """
        if self.Delta == 'fof':
            raise ValueError("FoF masses don't have an associated overdensity."
                             "Nor can they be translated into other masses")
        if self.Delta == 'vir':
            return self.get_Delta_vir(cosmo, a)
        return self.Delta

    def _get_Delta_m(self, cosmo, a):
        """ For SO-based mass definitions, this returns the corresponding
        value of Delta for a rho_matter-based definition.
        """
        delta = self.get_Delta(cosmo, a)
        if self.rho_type == 'matter':
            return delta
        om_this = cosmo.omega_x(a, self.rho_type)
        om_matt = cosmo.omega_x(a, 'matter')
        return delta * om_this / om_matt

    def get_mass(self, cosmo, R, a):
        """ Translates a halo radius into a mass

        .. math::
            M_\\Delta = \\frac{4 \\pi}{3} \\Delta\\,\\rho_X\\, R_\\Delta^3

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
            R (:obj:`float` or `array`): halo radius (:math:`{\\rm Mpc}`,
                physical, not comoving).
            a (:obj:`float`): scale factor.

        Returns:
            (:obj:`float` or `array`): halo mass in units of :math:`M_\\odot`.
        """
        R_use = np.atleast_1d(R)
        Delta = self.get_Delta(cosmo, a)
        M = 4.18879020479 * cosmo.rho_x(a, self.rho_type) * Delta * R_use**3
        if np.ndim(R) == 0:
            return M[0]
        return M

    def get_radius(self, cosmo, M, a):
        """ Translates a halo mass into a radius

        .. math::
            R_\\Delta = \\left(\\frac{3M_\\Delta}{4 \\pi
            \\rho_X\\,\\Delta}\\right)^{1/3}

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`):
                A Cosmology object.
            M (:obj:`float` or `array`): halo mass in units of
                :math:`M_\\odot`.
            a (:obj:`float`): scale factor.

        Returns:
            (:obj:`float` or `array`): halo radius in units of
            :math:`{\\rm Mpc}` (physical, not comoving).
        """
        M_use = np.atleast_1d(M)
        Delta = self.get_Delta(cosmo, a)
        R = (M_use / (4.18879020479 * Delta *
                      cosmo.rho_x(a, self.rho_type)))**(1./3.)
        if np.ndim(M) == 0:
            return R[0]
        return R

    @classmethod
    def from_name(cls, name):
        """ Return mass definition subclass from name string.

        Args:
            name (:obj:`str`):
                a mass definition name (e.g. ``'200m'`` for
                :math:`\\Delta=200` times the matter density).

        Returns:
            :class:`MassDef` instance corresponding to the input name.
        """
        MassDefName = f"MassDef{name.capitalize()}"
        if MassDefName in globals():
            # MassDef is defined in one of the implementations below.
            return globals()[MassDefName]
        parser = {"c": "critical", "m": "matter"}
        if len(name) < 2 or name[-1] not in parser:
            # Bogus input - can't parse it.
            raise ValueError("Could not parse mass definition string.")
        Delta, rho_type = name[:-1], parser[name[-1]]
        return cls(Delta, rho_type)

    @classmethod
    def create_instance(cls, input_):
        if isinstance(input_, cls):
            return input_
        else:
            return cls.from_name(input_)

    @classmethod
    def from_specs(cls, mass_def=None, *,
                   mass_function=None, halo_bias=None, concentration=None):
        """Instantiate mass definition and halo model ingredients.

        Unspecified halo model ingredients are ignored. ``mass_def`` is always
        instantiated.

        Args:
            mass_def (:obj:`MassDef`, :obj:`str` or :obj:`None`):
                Mass definition. If a string, instantiate from its name.
                If `None`, obtain the one from the first specified halo model
                ingredient.
            mass_function (:class:`~pyccl.halos.halo_model_base.MassFunc` or :obj:`str`):
                Mass function subclass. Strings are auto-instantiated using
                ``mass_def``. ``None`` values are ignored.
            halo_bias (:class:`~pyccl.halos.halo_model_base.HaloBias` or :obj:`str`):
                Halo bias subclass. Strings are auto-instantiated using
                ``mass_def``. ``None`` values are ignored.
            concentration (:class:`~pyccl.halos.halo_model_base.Concentration` or :obj:`str`):
                Concentration subclass. Strings are auto-instantiated using
                ``mass_def``. ``None`` values are ignored.

        Returns:
            Tuple of up to 4 elements.

            - mass_def : :class:`MassDef`
            - mass_function : :class:`~pyccl.halos.halo_model_base.MassFunc`, if specified
            - halo_bias : :class:`~pyccl.halos.halo_model_base.HaloBias`, if specified
            - concentration : :class:`~pyccl.halos.halo_model_base.Concentration`, if specified
        """ # noqa
        values = mass_function, halo_bias, concentration
        idx = [value is not None for value in values]

        # Filter only the specified ones.
        values = np.array(values)[idx]
        names = np.array(["mass_function", "halo_bias", "concentration"])[idx]
        Types = np.array([MassFunc, HaloBias, Concentration])[idx]

        # Sanity check.
        if mass_def is None:
            for name, value in zip(names, values):
                if isinstance(value, str):
                    raise ValueError(f"Need mass_def if {name} is str.")

        # Instantiate mass_def.
        if mass_def is not None:
            mass_def = cls.create_instance(mass_def)  # instantiate directly
        else:
            mass_def = values[0].mass_def  # use the one in HMIngredients

        # Instantiate halo model ingredients.
        out = []
        for name, value, Type in zip(names, values, Types):
            instance = Type.create_instance(value, mass_def=mass_def)
            out.append(instance)

        # Check mass definition consistency.
        if out and set([x.mass_def for x in out]) != set([mass_def]):
            raise ValueError("Inconsistent mass definitions.")

        return mass_def, *out


MassDef200m = MassDef(200, "matter")
MassDef200c = MassDef(200, "critical")
MassDef500c = MassDef(500, "critical")
MassDefVir = MassDef("vir", "critical")
MassDefFof = MassDef("fof", "matter")


def mass_translator(*, mass_in, mass_out, concentration):
    """Translate between mass definitions, assuming an NFW profile.

    Returns a function that can be used to translate between halo
    masses according to two different definitions.

    Args:
        mass_in (:class:`MassDef` or :obj:`str`): mass definition of the
            input mass.
        mass_out (:class:`MassDef` or :obj:`str`): mass definition of the
            output mass.
        concentration (:class:`~pyccl.halos.halo_model_base.Concentration` or :obj:`str`):
            concentration-mass relation to use for the mass conversion. It must
            be calibrated for masses using the ``mass_in`` definition.

    Returns:
        Function that ranslates between two masses. The returned function
        ``f`` can be called as: ``f(cosmo, M, a)``, where
        ``cosmo`` is a :class:`~pyccl.cosmology.Cosmology` object, ``M``
        is a mass (or array of masses), and ``a`` is a scale factor.

    """ # noqa

    mass_in = MassDef.create_instance(mass_in)
    mass_out = MassDef.create_instance(mass_out)
    concentration = Concentration.create_instance(concentration,
                                                  mass_def=mass_in)
    if concentration.mass_def != mass_in:
        raise ValueError("mass_def of concentration doesn't match mass_in")

    def translate(cosmo, M, a):
        if mass_in == mass_out:
            return M

        c_in = concentration(cosmo, M, a)
        Om_in = cosmo.omega_x(a, mass_in.rho_type)
        D_in = mass_in.get_Delta(cosmo, a) * Om_in
        R_in = mass_in.get_radius(cosmo, M, a)

        Om_out = cosmo.omega_x(a, mass_out.rho_type)
        D_out = mass_out.get_Delta(cosmo, a) * Om_out
        c_out = convert_concentration(
            cosmo, c_old=c_in, Delta_old=D_in, Delta_new=D_out)
        R_out = R_in * c_out/c_in
        return mass_out.get_mass(cosmo, R_out, a)

    return translate



__all__ = ("translate_IA_norm", "PTTracer", "PTMatterTracer",
           "PTNumberCountsTracer", "PTIntrinsicAlignmentTracer",)

from abc import abstractmethod

import numpy as np
from scipy.interpolate import interp1d

from .. import CCLAutoRepr, physical_constants
from ..pyutils import _check_array_params


def translate_IA_norm(cosmo, *, z, a1=1.0, a1delta=None, a2=None,
                      Om_m2_for_c2=False, Om_m_fid=0.3):
    """
    Function to convert from :math:`A_{ia}` values to :math:`c_{ia}` values,
    for the intrinsic alignment bias parameters using the standard
    convention of `Blazek et al. 2019 <https://arxiv.org/abs/1708.09247>`_
    or the variant used by the Dark Energy Survey analysis.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): cosmology object.
        z (:obj:`float` or `array`): z value(s) where amplitude is evaluated.
        a1 (:obj:`float` or `array`): IA :math:`A_1` at input z values.
        a1delta (:obj:`float` or `array`): IA :math:`A_{1\\delta}` at input
            z values.
        a2 (:obj:`float` or `array`): IA :math:`A_2` at input z values.
        Om_m2_for_c2 (:obj:`bool`): True to use the Blazek et al. 2019
            convention of :math:`\\Omega_m^2` scaling.
        Om_m_fid (:obj:`float`): Value for Blazek et al. 2019 scaling.

    Returns:
        Tuple of IA bias parameters

        - c1 (:obj:`float` or `array`): IA :math:`C_1` at input z values.
        - c1delta (:obj:`float` or `array`): IA :math:`C_{1\\delta}` at
          input z values.
        - c2 (:obj:`float` or `array`): IA :math:`C_2` at input z values.
    """

    Om_m = cosmo['Omega_m']
    rho_crit = physical_constants.RHO_CRITICAL
    c1 = c1delta = c2 = None
    gz = cosmo.growth_factor(1./(1+z))

    if a1 is not None:
        c1 = -1*a1*5e-14*rho_crit*Om_m/gz

    if a1delta is not None:
        c1delta = -1*a1delta*5e-14*rho_crit*Om_m/gz

    if a2 is not None:
        if Om_m2_for_c2:  # Blazek2019 convention
            c2 = a2*5*5e-14*rho_crit*Om_m**2/(Om_m_fid*gz**2)
        else:  # DES convention
            c2 = a2*5*5e-14*rho_crit*Om_m/(gz**2)

    return c1, c1delta, c2


class PTTracer(CCLAutoRepr):
    """PTTracers contain the information necessary to describe the
    perturbative, non-linear inhomogeneities associated with
    different physical quantities.

    In essence their main function is to store a set of
    redshift-dependent functions (e.g. perturbation theory biases)
    needed in a perturbation theory framework to provide N-point
    correlations.
    """
    __repr_attrs__ = __eq_attrs__ = ('type', 'biases')

    def __init__(self):
        self.biases = {}
        pass

    @property
    @abstractmethod
    def type(self):
        """String defining tracer type (``'M'``, ``'NC'`` and
        ``'IA'`` supported).
        """

    def get_bias(self, bias_name, z):
        """Get the value of one of the bias functions at a given
        redshift.

        Args:
            bias_name (:obj:`str`): name of the bias function to return.
            z (:obj:`float` or `array`): redshift.

        Returns:
            (:obj:`float` or `array`): bias value at the input redshifts.
        """
        if bias_name not in self.biases:
            raise KeyError(f"Bias {bias_name} not included in this tracer")
        return self.biases[bias_name](z)

    def _get_bias_function(self, b):
        # If None, assume it's zero
        if b is None:
            b = 0

        # If it's a scalar, then assume it's a constant function
        if np.ndim(b) == 0:
            def _const(z):
                if np.ndim(z) == 0:
                    return b
                else:
                    return b * np.ones_like(z)

            return _const
        else:  # Otherwise interpolate
            z, b = _check_array_params(b)
            return interp1d(z, b, bounds_error=False,
                            fill_value=b[-1])


class PTMatterTracer(PTTracer):
    """:class:`PTTracer` representing matter fluctuations.
    """
    type = 'M'

    def __init__(self):
        self.biases = {}


class PTNumberCountsTracer(PTTracer):
    """:class:`PTTracer` representing number count fluctuations.
    This is described by 1st and 2nd-order biases and
    a tidal field bias. These are provided as floating
    point numbers or tuples of `(reshift,bias)` arrays.
    If a number is provided, a constant bias is assumed.
    If ``None``, a bias of zero is assumed.

    Args:
        b1 (:obj:`float` or :obj:`tuple`): a single number or a
            tuple of arrays ``(z, b(z))`` giving the first-order
            bias.
        b2 (:obj:`float` or :obj:`tuple`): as above for the
            second-order bias.
        bs (:obj:`float` or :obj:`tuple`): as above for the
            tidal bias.
        b3nl (:obj:`float` or :obj:`tuple`): as above for the
            third-order bias.
        bk2 (:obj:`float` or :obj:`tuple`): as above for the
            non-local bias.
    """
    type = 'NC'

    def __init__(self, b1, b2=None, bs=None, b3nl=None, bk2=None):
        self.biases = {}

        # Initialize b1
        self.biases['b1'] = self._get_bias_function(b1)
        # Initialize b2
        self.biases['b2'] = self._get_bias_function(b2)
        # Initialize bs
        self.biases['bs'] = self._get_bias_function(bs)
        # Initialize b3nl
        self.biases['b3nl'] = self._get_bias_function(b3nl)
        # Initialize bk2
        self.biases['bk2'] = self._get_bias_function(bk2)

    @property
    def b1(self):
        """Internal first-order bias function.
        """
        return self.biases['b1']

    @property
    def b2(self):
        """Internal second-order bias function.
        """
        return self.biases['b2']

    @property
    def bs(self):
        """Internal tidal bias function.
        """
        return self.biases['bs']

    @property
    def b3nl(self):
        """Internal third-order bias function.
        """
        return self.biases['b3nl']

    @property
    def bk2(self):
        """Internal non-local bias function.
        """
        return self.biases['bk2']


class PTIntrinsicAlignmentTracer(PTTracer):
    """:class:`PTTracer` representing intrinsic alignments.
    This is described by 1st and 2nd-order alignment biases
    and an overdensity bias. These are provided as floating
    point numbers or tuples of (reshift,bias) arrays.
    If a number is provided, a constant bias is assumed.
    If ``None``, a bias of zero is assumed.

    Args:
        c1 (:obj:`float` or :obj:`tuple`): a single number or a
            tuple of arrays ``(z, c1(z))`` giving the first-order
            alignment bias :math:`C_1`.
        c2 (:obj:`float` or :obj:`tuple`): as above for the
            second-order alignment bias :math:`C_2`.
        cdelta (:obj:`float` or :obj:`tuple`): as above for the
            overdensity bias :math:`C_{1\\delta}`.
    """
    type = 'IA'

    def __init__(self, c1, c2=None, cdelta=None):

        self.biases = {}

        # Initialize c1
        self.biases['c1'] = self._get_bias_function(c1)
        # Initialize c2
        self.biases['c2'] = self._get_bias_function(c2)
        # Initialize cdelta
        self.biases['cdelta'] = self._get_bias_function(cdelta)

    @property
    def c1(self):
        """Internal first-order bias function.
        """
        return self.biases['c1']

    @property
    def c2(self):
        """Internal second-order bias function.
        """
        return self.biases['c2']

    @property
    def cdelta(self):
        """Internal overdensity bias function.
        """
        return self.biases['cdelta']



__all__ = ("MassFuncDespali16",)

import numpy as np

from . import MassFunc, get_delta_c


class MassFuncDespali16(MassFunc):
    """Implements the mass function of `Despali et al. 2016
    <https://arxiv.org/abs/1507.05627>`_. This parametrization accepts
    any S.O. masses.


    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`):
            a mass definition object, or a name string.
        mass_def_strict (:obj:`bool`): if ``False``, consistency of the mass
            definition will be ignored.
        ellipsoidal (:obj:`bool`): use the ellipsoidal parametrization.
    """
    __repr_attrs__ = __eq_attrs__ = ("mass_def", "mass_def_strict",
                                     "ellipsoidal",)
    name = 'Despali16'

    def __init__(self, *,
                 mass_def="200m",
                 mass_def_strict=True,
                 ellipsoidal=False):
        self.ellipsoidal = ellipsoidal
        super().__init__(mass_def=mass_def, mass_def_strict=mass_def_strict)

    def _check_mass_def_strict(self, mass_def):
        # True for FoF since Despali16 is not defined for this mass def.
        return mass_def.Delta == "fof"

    def _setup(self):
        # key: ellipsoidal
        vals = {True: (0.3953, -0.1768, 0.7057, 0.2125, 0.3268,
                       0.2206, 0.1937, -0.04570),
                False: (0.3292, -0.1362, 0.7665, 0.2263, 0.4332,
                        0.2488, 0.2554, -0.1151)}

        A0, A1, a0, a1, a2, p0, p1, p2 = vals[self.ellipsoidal]
        coeffs = [[A1, A0], [a2, a1, a0], [p2, p2, p0]]
        self.poly_A, self.poly_a, self.poly_p = map(np.poly1d, coeffs)

    def _get_fsigma(self, cosmo, sigM, a, lnM):
        delta_c = get_delta_c(cosmo, a, 'NakamuraSuto97')

        Dv = self.mass_def.get_Delta_vir(cosmo, a)
        x = np.log10(self.mass_def.get_Delta(cosmo, a) / Dv)

        A, a, p = self.poly_A(x), self.poly_a(x), self.poly_p(x)

        nu_p = a * (delta_c/sigM)**2
        return 2.0 * A * np.sqrt(nu_p / 2.0 / np.pi) * (
            np.exp(-0.5 * nu_p) * (1.0 + nu_p**-p))



import warnings
import pyccl
import numpy as np
from .hod import HaloProfileHOD


__all__ = ("SatelliteShearHOD",)


class SatelliteShearHOD(HaloProfileHOD):
    """ Halo HOD class that calculates the satellite galaxy intrinsic shear
    field in real and fourier space, according to `Fortuna et al. 2021.
    <https://arxiv.org/abs/2003.02700>`_.
    Can be used to compute halo model-based intrinsic alignment
    (angular) power spectra.

    The satellite intrinsic shear profile in real space is assumed to be

    .. math::
        \\gamma^I(r)=a_{1\\mathrm{h}}\\left(\\frac{r}{r_\\mathrm{vir}}
        \\right)^b \\sin^b\\theta,

    where :math:`a_{1\\mathrm{h}}` is the amplitude of intrinsic alignments on
    the 1-halo scale, :math:`b` the index defining the radial dependence
    and :math:`\\theta` the angle defining the projection of the
    semi-major axis of the galaxy along the line of sight.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`):
            a mass definition object, or a name string.
        concentration (:obj:`~pyccl.halos.halo_model_base.Concentration`):
            concentration-mass relation to use with this profile.
        a1h (:obj:`float`): Amplitude of the satellite intrinsic shear profile.
        b (:obj:`float`): Power-law index of the satellite intrinsic shear profile.
            If zero, the profile is assumed to be constant inside the halo.
        lmax (:obj:`int`): Maximum multipole to be summed in the plane-wave expansion
            (Eq. (C1) in `Fortuna et al. 2021
            <https://arxiv.org/abs/2003.02700>`_, default=6).
        log10Mmin_0 (:obj:`float`): offset parameter for
            :math:`\\log_{10}M_{\\rm min}`.
        log10Mmin_p (:obj:`float`): tilt parameter for
            :math:`\\log_{10}M_{\\rm min}`.
        siglnM_0 (:obj:`float`): offset parameter for
            :math:`\\sigma_{{\\rm ln}M}`.
        siglnM_p (:obj:`float`): tilt parameter for
            :math:`\\sigma_{{\\rm ln}M}`.
        log10M0_0 (:obj:`float`): offset parameter for
            :math:`\\log_{10}M_0`.
        log10M0_p (:obj:`float`): tilt parameter for
            :math:`\\log_{10}M_0`.
        log10M1_0 (:obj:`float`): offset parameter for
            :math:`\\log_{10}M_1`.
        log10M1_p (:obj:`float`): tilt parameter for
            :math:`\\log_{10}M_1`.
        alpha_0 (:obj:`float`): offset parameter for
            :math:`\\alpha`.
        alpha_p (:obj:`float`): tilt parameter for
            :math:`\\alpha`.
        bg_0 (:obj:`float`): offset parameter for
            :math:`\\beta_g`.
        bg_p (:obj:`float`): tilt parameter for
            :math:`\\beta_g`.
        bmax_0 (:obj:`float`): offset parameter for
            :math:`\\beta_{\\rm max}`.
        bmax_p (:obj:`float`): tilt parameter for
            :math:`\\beta_{\\rm max}`.
        a_pivot (:obj:`float`): pivot scale factor :math:`a_*`.
        ns_independent (:obj:`bool`): drop requirement to only form
            satellites when centrals are present.
        integration_method (:obj:`str`): Method used to obtain the fourier transform
            of the profile. Can be ``'FFTLog'``, ``'simpson'`` or
            ``'spline'``.
        rmin (:obj:`float`): For ``'simpson'`` or ``'spline'`` integration, minimum value of
            physical radius used to carry out the radial integral (in Mpc).
        N_r (:obj:`int`): For ``'simpson'`` or ``'spline'`` integration, number of points
            to be used when sampling the radial integral (in log space).
        N_jn (:obj:`int`): For ``'simpson'`` or ``'spline'`` integration, number of points
            to be used when sampling the spherical Bessel functions, that are
            later used to interpolate. Interpolating the Bessel functions
            increases the speed of the computations compared to explicitly
            evaluating them, without significant loss of accuracy.
    """ # noqa
    __repr_attrs__ = __eq_attrs__ = (
        "a1h", "b", "lmax", "integration_method",
        "log10Mmin_0", "log10Mmin_p", "siglnM_0", "siglnM_p", "log10M0_0",
        "log10M0_p", "log10M1_0", "log10M1_p", "alpha_0", "alpha_p",
        "bg_0", "bg_p", "bmax_0", "bmax_p", "a_pivot", "ns_independent",
        "rmin", "N_r", "N_jn", "concentration", "integration_method",
        "precision_fftlog",)

    def __init__(self, *, mass_def, concentration, a1h=0.001, b=-2,
                 lmax=6, log10Mmin_0=12., log10Mmin_p=0., siglnM_0=0.4,
                 siglnM_p=0., log10M0_0=7., log10M0_p=0.,
                 log10M1_0=13.3, log10M1_p=0., alpha_0=1.,
                 alpha_p=0., bg_0=1., bg_p=0., bmax_0=1., bmax_p=0.,
                 a_pivot=1., ns_independent=False,
                 integration_method='FFTLog', rmin=0.001, N_r=512,
                 N_jn=10000):
        if lmax >= 13:
            lmax = 12
            warnings.warn(
                'Maximum l provided too high. Using lmax=12.',
                category=pyccl.CCLWarning)
        elif lmax < 2:
            lmax = 2
            warnings.warn(
                'Maximum l provided too low. Using lmax=2.',
                category=pyccl.CCLWarning)
        self.a1h = a1h
        self.b = b
        if integration_method not in ['FFTLog',
                                      'simpson',
                                      'spline']:
            raise ValueError("Integration method provided not "
                             "supported. Use `FFTLog`, `simpson`, "
                             "or `spline`.")
        self.integration_method = integration_method
        # If lmax is odd, make it even number (odd l contributions are zero).
        if not (lmax % 2 == 0):
            lmax = lmax-1
        self.lmax = lmax
        # Hard-code for most common cases (b=0, b=-2) to gain speed (~1.3sec).
        if self.b == 0:
            self._angular_fl = np.array([2.77582637, -0.19276603,
                                         0.04743899, -0.01779024,
                                         0.00832446, -0.00447308])\
                .reshape((6, 1))
        elif self.b == -2:
            self._angular_fl = np.array([4.71238898, -2.61799389,
                                         2.06167032, -1.76714666,
                                         1.57488973, -1.43581368])\
                .reshape((6, 1))
        else:
            self._angular_fl = np.array([self._fl(l, b=self.b)
                                         for l in range(2, self.lmax+1, 2)])\
                .reshape(self.lmax//2, 1)
        self.concentration = concentration
        self.rmin = rmin
        self.N_r = N_r
        self.N_jn = N_jn
        super().__init__(mass_def=mass_def,
                         concentration=concentration,
                         log10Mmin_0=log10Mmin_0,
                         log10Mmin_p=log10Mmin_p,
                         siglnM_0=siglnM_0,
                         siglnM_p=siglnM_p,
                         log10M0_0=log10M0_0,
                         log10M0_p=log10M0_p,
                         log10M1_0=log10M1_0,
                         log10M1_p=log10M1_p,
                         fc_0=0.0, fc_p=0.0,
                         alpha_0=alpha_0,
                         alpha_p=alpha_p,
                         bg_0=bg_0,
                         bg_p=bg_p,
                         bmax_0=bmax_0,
                         bmax_p=bmax_p,
                         a_pivot=a_pivot,
                         ns_independent=ns_independent)
        self.update_precision_fftlog(padding_lo_fftlog=1E-2,
                                     padding_hi_fftlog=1E3,
                                     n_per_decade=350,
                                     plaw_fourier=-3.7)

    def update_parameters(self, *, a1h=None, b=None, lmax=None,
                          log10Mmin_0=None, log10Mmin_p=None,
                          siglnM_0=None, siglnM_p=None,
                          log10M0_0=None, log10M0_p=None,
                          log10M1_0=None, log10M1_p=None,
                          alpha_0=None, alpha_p=None,
                          bg_0=None, bg_p=None,
                          bmax_0=None, bmax_p=None,
                          a_pivot=None, ns_independent=None,
                          rmin=None, N_r=None, N_jn=None):
        """ Update any of the parameters associated with
        this profile. Any parameter set to `None` won't be updated.

        Args:
            a1h (:obj:`float`): Amplitude of the satellite intrinsic shear profile.
            b (:obj:`float`): Power-law index of the satellite intrinsic shear
                profile. If zero, the profile is assumed to be constant inside
                the halo.
            lmax (:obj:`int`): Maximum multipole to be summed in the plane-wave
                expansion.
            log10Mmin_0 (:obj:`float`): offset parameter for
                :math:`\\log_{10}M_{\\rm min}`.
            log10Mmin_p (:obj:`float`): tilt parameter for
                :math:`\\log_{10}M_{\\rm min}`.
            siglnM_0 (:obj:`float`): offset parameter for
                :math:`\\sigma_{{\\rm ln}M}`.
            siglnM_p (:obj:`float`): tilt parameter for
                :math:`\\sigma_{{\\rm ln}M}`.
            log10M0_0 (:obj:`float`): offset parameter for
                :math:`\\log_{10}M_0`.
            log10M0_p (:obj:`float`): tilt parameter for
                :math:`\\log_{10}M_0`.
            log10M1_0 (:obj:`float`): offset parameter for
                :math:`\\log_{10}M_1`.
            log10M1_p (:obj:`float`): tilt parameter for
                :math:`\\log_{10}M_1`.
            alpha_0 (:obj:`float`): offset parameter for
                :math:`\\alpha`.
            alpha_p (:obj:`float`): tilt parameter for
                :math:`\\alpha`.
            bg_0 (:obj:`float`): offset parameter for
                :math:`\\beta_g`.
            bg_p (:obj:`float`): tilt parameter for
                :math:`\\beta_g`.
            bmax_0 (:obj:`float`): offset parameter for
                :math:`\\beta_{\\rm max}`.
            bmax_p (:obj:`float`): tilt parameter for
                :math:`\\beta_{\\rm max}`.
            a_pivot (:obj:`float`): pivot scale factor :math:`a_*`.
            ns_independent (:obj:`bool`): drop requirement to only form
                satellites when centrals are present
            rmin (:obj:`float`): For `simpson` or `spline` integration, minimum value of
                physical radius used to carry out the radial integral (in Mpc).
            N_r (:obj:`int`): For `simpson` or `spline` integration, number of points to be
                used when sampling the radial integral (in logarithmic space).
            N_jn (:obj:`int`): For `simpson` or `spline` integration, number of points to
                be used when sampling the spherical Bessel functions, that are
                later used to interpolate. Interpolating the Bessel functions
                increases the speed of the computations compared to explicitly
                evaluating them, without significant loss of accuracy.
        """ # noqa
        if a1h is not None:
            self.a1h = a1h
        if b is not None:
            self.b = b
        if lmax is not None:
            self.lmax = lmax
        if log10Mmin_0 is not None:
            self.log10Mmin_0 = log10Mmin_0
        if log10Mmin_p is not None:
            self.log10Mmin_p = log10Mmin_p
        if log10M0_0 is not None:
            self.log10M0_0 = log10M0_0
        if log10M0_p is not None:
            self.log10M0_p = log10M0_p
        if log10M1_0 is not None:
            self.log10M1_0 = log10M1_0
        if log10M1_p is not None:
            self.log10M1_p = log10M1_p
        if siglnM_0 is not None:
            self.siglnM_0 = siglnM_0
        if siglnM_p is not None:
            self.siglnM_p = siglnM_p
        if alpha_0 is not None:
            self.alpha_0 = alpha_0
        if alpha_p is not None:
            self.alpha_p = alpha_p
        if bg_0 is not None:
            self.bg_0 = bg_0
        if bg_p is not None:
            self.bg_p = bg_p
        if bmax_0 is not None:
            self.bmax_0 = bmax_0
        if bmax_p is not None:
            self.bmax_p = bmax_p
        if a_pivot is not None:
            self.a_pivot = a_pivot
        if ns_independent is not None:
            self.ns_independent = ns_independent
        if rmin is not None:
            self.rmin = rmin
        if N_r is not None:
            self.N_r = N_r
        if N_jn is not None:
            self.N_jn = N_jn

    def _I_integral(self, a, b):
        '''
        Computes the integral

        .. math::
            I(a,b) = \\int_{-1}^1 \\mathrm{d}x (1-x^2)^{a/2}x^b =
            \\frac{((-1)^b+1)\\Gamma(a+1)\\Gamma\\left
            \\frac{b+1}{2}\\right}
            {2\\Gamma\\left(a+\\frac{b}{2}+\\frac{3}{2}\\right}.
        '''
        from scipy.special import gamma
        return (1+(-1)**b)*gamma(a/2+1)*gamma((b+1)/2)/(2*gamma(a/2+b/2+3/2))

    def _fl(self, l, thk=np.pi / 2, phik=None, b=-2):
        '''
        Computes the angular part of the satellite intrinsic shear field,
        Eq. (C8) in `Fortuna et al. 2021 <https://arxiv.org/abs/2003.02700>`_.
        '''
        from scipy.special import binom
        gj = np.array([0, 0, np.pi / 2, 0, np.pi / 2, 0, 15 * np.pi / 32,
                       0, 7 * np.pi / 16, 0, 105 * np.pi / 256, 0,
                       99 * np.pi / 256])
        l_sum = 0.
        if b == 0:
            var1_add = 1
        else:
            var1_add = b
        for m in range(0, l + 1):
            m_sum = 0.
            for j in range(0, m + 1):
                m_sum += (binom(m, j) * gj[j] *
                          self._I_integral(j + var1_add, m - j) *
                          np.sin(thk) ** j * np.cos(thk) ** (m - j))
            l_sum += binom(l, m) * binom((l + m - 1) / 2, l) * m_sum
        if phik is not None:
            l_sum *= np.exp(1j * 2 * phik)
        return 2 ** l * l_sum

    def get_normalization(self, cosmo, a, *, hmc):
        """Returns the normalization of this profile, which is the
        mean galaxy number density.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
            a (:obj:`float`): scale factor.
            hmc (:class:`~pyccl.halos.halo_model.HMCalculator`): a halo
                model calculator object.

        Returns:
            float: normalization factor of this profile.
        """
        def integ(M):
            Nc = self._Nc(M, a)
            Ns = self._Ns(M, a)
            if self.ns_independent:
                return Nc+Ns
            return Nc*(1+Ns)
        return hmc.integrate_over_massfunc(integ, cosmo, a)

    def gamma_I(self, r, r_vir):
        '''
        Returns the intrinsic satellite shear,

        .. math::
            \\gamma^I(r)=a_{1\\mathrm{h}}
            \\left(\\frac{r}{r_\\mathrm{vir}}\\right)^b.

        If :math:`b` is 0, then only the value of the amplitude
        :math:`a_\\mathrm{1h}` is returned. In addition, according to
        `Fortuna et al. 2021 <https://arxiv.org/abs/2003.02700>`_, we
        use a constant value of 0.06 Mpc for
        :math:`r<0.06` Mpc and set a maximum of 0.3 for :math:`\\gamma^I(r)`.
        '''
        if self.b == 0:
            return self.a1h
        else:
            r_use = np.copy(np.atleast_1d(r))
            r_use[r_use < 0.06] = 0.06
            if np.ndim(r_vir == 1):
                r_vir = r_vir.reshape(len(r_vir), 1)
            # Do not output value higher than 0.3
            gamma_out = self.a1h * (r_use/r_vir) ** self.b
            gamma_out[gamma_out > 0.3] = 0.3
            return gamma_out

    def _real(self, cosmo, r, M, a):
        '''
        Returns the real part of the satellite intrinsic shear field,

        .. math::
            \\gamma^I(r) u(r|M),

        with :math:`u` being the halo density profile divided by its mass.
        For now, it assumes a NFW profile.
        '''
        M_use = np.atleast_1d(M)
        r_use = np.atleast_1d(r)

        rvir = self.mass_def.get_radius(cosmo, M_use, a) / a
        # Density profile from HOD class - truncated NFW
        u = self._usat_real(cosmo, r_use, M_use, a)
        prof = self.gamma_I(r_use, rvir) * u

        if np.ndim(r) == 0:
            prof = np.squeeze(prof, axis=-1)
        if np.ndim(M) == 0:
            prof = np.squeeze(prof, axis=0)

        return prof

    def _usat_fourier(self, cosmo, k, M, a):
        '''
        Returns the fourier transform of the satellite intrinsic shear field.
        The density profile of the halo is assumed to be a truncated NFW
        profile and the radial integral (in the case of 'simpson' or 'spline'
        method) is evaluated up to the virial radius.
        '''
        M_use = np.atleast_1d(M)
        k_use = np.atleast_1d(k)
        l_arr = np.arange(2, self.lmax+1, 2, dtype='int32')
        if not self.integration_method == 'FFTLog':
            # Define the r-integral sampling and the sph. bessel
            # function sampling. The bessel function will be sampled
            # and interpolated to gain speed.
            r_use = np.linspace(self.rmin,
                                self.mass_def.get_radius(cosmo, M_use, a) / a,
                                self.N_r).T
            x_jn = np.geomspace(k_use.min() * r_use.min(),
                                k_use.max() * r_use.max(),
                                self.N_jn)
            jn = np.empty(shape=(len(l_arr), len(x_jn)))
        prof = np.zeros(shape=(len(M_use), len(k_use)))
        # Loop over all multipoles:
        for j, l in enumerate(l_arr):
            if self.integration_method == 'FFTLog':
                prof += (self._fftlog_wrap(cosmo, k_use, M_use, a,
                                           ell=int(l), fourier_out=True)
                         * (1j**l).real * (2*l+1) * self._angular_fl[j]
                         / (4*np.pi))
            else:
                from scipy.special import spherical_jn
                jn[j] = spherical_jn(l_arr[j], x_jn)
                k_dot_r = np.multiply.outer(k_use, r_use)
                jn_interp = np.interp(k_dot_r, x_jn, jn[j])
                integrand = (r_use**2 * jn_interp *
                             self._real(cosmo, r_use, M_use, a))
                if self.integration_method == 'simpson':
                    from scipy.integrate import simpson
                    for i, M_i in enumerate(M_use):
                        prof[i, :] += (simpson(integrand[:, i, :],
                                               r_use[i, :]).T
                                       * (1j**l).real * (2*l+1)
                                       * self._angular_fl[j])
                elif self.integration_method == 'spline':
                    from pyccl.pyutils import _spline_integrate
                    for i, M_i in enumerate(M_use):
                        prof[i, :] += (_spline_integrate(r_use[i, :],
                                                         integrand[:, i, :],
                                                         r_use[i, 0],
                                                         r_use[i, -1]).T
                                       * (1j**l).real*(2*l+1)
                                       * self._angular_fl[j])

        if np.ndim(k) == 0:
            prof = np.squeeze(prof, axis=-1)
        if np.ndim(M) == 0:
            prof = np.squeeze(prof, axis=0)

        return prof



import numpy as np
import pytest
import pyccl as ccl
from pyccl.modified_gravity import MuSigmaMG


POWER_MG_TOL = 1e-2


@pytest.mark.parametrize('model', list(range(5)))
def test_power_mg(model):
    mu_0 = [0., 0.1, -0.1, 0.1, -0.1]
    sigma_0 = [0., 0.1, -0.1, -0.1, 0.1]

    cosmo = ccl.Cosmology(
        Omega_c=0.25,
        Omega_b=0.05,
        h=0.7,
        A_s=2.1e-9,
        n_s=0.96,
        Neff=3.046,
        Omega_k=0,
        mg_parametrization=MuSigmaMG(
            mu_0=mu_0[model],
            sigma_0=sigma_0[model]),
        matter_power_spectrum='linear',
        transfer_function='boltzmann_class')

    data = np.loadtxt("./benchmarks/data/model%d_pk_MG.dat" % model)

    a = 1
    k = data[:, 0] * cosmo['h']
    pk = data[:, 1] / (cosmo['h']**3)
    pk_ccl = ccl.linear_matter_power(cosmo, k, a)
    err = np.abs(pk_ccl/pk - 1)

    if model == 0:
        assert np.allclose(err, 0, rtol=0, atol=POWER_MG_TOL)
    else:
        msk = data[:, 0] > 5e-3
        assert np.allclose(err[msk], 0, rtol=0, atol=POWER_MG_TOL)



import pickle
import tempfile
import pytest
import numpy as np
import pyccl as ccl
import copy
import warnings
from .test_cclobject import check_eq_repr_hash


def test_Cosmology_eq_repr_hash():
    # Test eq, repr, hash for Cosmology and CosmologyCalculator.
    # 1. Using a complicated Cosmology object.
    extras = {"camb": {"halofit_version": "mead2020", "HMCode_logT_AGN": 7.8}}
    kwargs = {"transfer_function": "bbks",
              "matter_power_spectrum": "linear",
              "extra_parameters": extras}
    COSMO1 = ccl.CosmologyVanillaLCDM(**kwargs)
    COSMO2 = ccl.CosmologyVanillaLCDM(**kwargs)
    assert check_eq_repr_hash(COSMO1, COSMO2)

    # 2. Now make a copy and change it.
    kwargs = copy.deepcopy(kwargs)
    kwargs["extra_parameters"]["camb"]["halofit_version"] = "mead2020_feedback"
    COSMO3 = ccl.CosmologyVanillaLCDM(**kwargs)
    assert check_eq_repr_hash(COSMO1, COSMO3, equal=False)

    # 3. Using a CosmologyCalculator.
    COSMO1.compute_linear_power()
    a_arr, lk_arr, pk_arr = COSMO1.get_linear_power().get_spline_arrays()
    pk_linear = {"a": a_arr,
                 "k": np.exp(lk_arr),
                 "delta_matter:delta_matter": pk_arr}
    COSMO4 = ccl.CosmologyCalculator(
        Omega_c=0.25, Omega_b=0.05, h=0.67, n_s=0.96, sigma8=0.81,
        pk_linear=pk_linear, pk_nonlin=pk_linear)
    COSMO5 = ccl.CosmologyCalculator(
        Omega_c=0.25, Omega_b=0.05, h=0.67, n_s=0.96, sigma8=0.81,
        pk_linear=pk_linear, pk_nonlin=pk_linear)
    assert check_eq_repr_hash(COSMO4, COSMO5)

    pk_linear = {"a": a_arr,
                 "k": np.exp(lk_arr),
                 "delta_matter:delta_matter": 2*pk_arr}
    COSMO6 = ccl.CosmologyCalculator(
        Omega_c=0.25, Omega_b=0.05, h=0.67, n_s=0.96, sigma8=0.81,
        pk_linear=pk_linear, pk_nonlin=pk_linear)
    assert check_eq_repr_hash(COSMO4, COSMO6, equal=False)


def test_cosmo_methods():
    """ Check that all pyccl functions that take cosmo
    as their first argument are methods of the Cosmology object.
    """
    from inspect import getmembers, isfunction, signature
    from pyccl import background, boltzmann, \
        cells, correlations, covariances, neutrinos, \
        pk2d, power, tk3d, tracers, halos, nl_pt
    cosmo = ccl.CosmologyVanillaLCDM()
    subs = [background, boltzmann, cells, correlations, covariances,
            neutrinos, pk2d, power, tk3d, tracers, halos, nl_pt]
    funcs = [getmembers(sub, isfunction) for sub in subs]
    funcs = [func for sub in funcs for func in sub]
    for name, func in funcs:
        if name.startswith("_"):  # no private functions
            continue
        pars = signature(func).parameters
        if pars and list(pars)[0] == "cosmo":
            _ = getattr(cosmo, name)

    # quantitative
    assert ccl.sigma8(cosmo) == cosmo.sigma8()
    assert ccl.rho_x(cosmo, 1., "matter", is_comoving=False) == \
        cosmo.rho_x(1., "matter", is_comoving=False)
    assert ccl.get_camb_pk_lin(cosmo)(1., 1., cosmo) == \
        cosmo.get_camb_pk_lin()(1., 1., cosmo)
    prof = ccl.halos.HaloProfilePressureGNFW(mass_def="200m")
    hmf = ccl.halos.MassFuncTinker08(mass_def="200m")
    hbf = ccl.halos.HaloBiasTinker10(mass_def="200m")
    hmc = ccl.halos.HMCalculator(mass_function=hmf, halo_bias=hbf,
                                 mass_def="200m")
    assert ccl.halos.halomod_power_spectrum(cosmo, hmc, 1., 1., prof) == \
        cosmo.halomod_power_spectrum(hmc, 1., 1., prof)


def test_cosmology_critical_init():
    cosmo = ccl.Cosmology(
        Omega_c=0.25,
        Omega_b=0.05,
        h=0.7,
        sigma8=0.8,
        n_s=0.96,
        Neff=0,
        m_nu=0.0,
        w0=-1.0,
        wa=0.0,
        mass_split='normal',
        Omega_g=0,
        Omega_k=0)
    assert np.allclose(cosmo.cosmo.data.growth0, 1)


def test_cosmology_As_sigma8_populates():
    # Check that cosmo.sigma8() pupulates sigma8 if it is missing.
    cosmo = ccl.Cosmology(Omega_c=0.265, Omega_b=0.045, h=0.675,
                          n_s=0.965, A_s=2e-9)
    assert np.isnan(cosmo["sigma8"])
    cosmo.sigma8()
    assert cosmo["sigma8"] == cosmo.sigma8()


def test_cosmology_init():
    """
    Check that Cosmology objects can only be constructed in a valid way.
    """
    # Make sure error raised if invalid transfer/power spectrum etc. passed
    with pytest.raises(KeyError):
        ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, h=0.7, A_s=2.1e-9, n_s=0.96,
                      matter_power_spectrum='x')
    with pytest.raises(KeyError):
        ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, h=0.7, A_s=2.1e-9, n_s=0.96,
                      transfer_function='x')
    with pytest.raises(ValueError):
        ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, h=0.7, A_s=2.1e-9, n_s=0.96,
                      m_nu=np.array([0.1, 0.1, 0.1, 0.1]))
    with pytest.raises(ValueError):
        ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, h=0.7, A_s=2.1e-9, n_s=0.96,
                      m_nu=ccl)


def test_cosmology_output():
    """
    Check that status messages and other output from Cosmology() object works
    correctly.
    """
    # Create test cosmology object
    cosmo = ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, h=0.7, A_s=2.1e-9,
                          n_s=0.96)

    # Return and print status messages
    assert cosmo.cosmo.status == 0

    # Test status methods for different precomputable quantities
    assert not cosmo.has_distances
    assert not cosmo.has_growth
    assert not cosmo.has_linear_power
    assert not cosmo.has_nonlin_power
    assert not cosmo.has_sigma

    cosmo.compute_distances()
    cosmo.compute_growth()
    cosmo.compute_linear_power()
    cosmo.compute_nonlin_power()
    cosmo.compute_sigma()

    assert cosmo.has_distances
    assert cosmo.has_growth
    assert cosmo.has_linear_power
    assert cosmo.has_nonlin_power
    assert cosmo.has_sigma


def test_cosmology_pickles():
    """Check that a Cosmology object pickles."""
    cosmo = ccl.Cosmology(
        Omega_c=0.25, Omega_b=0.05, h=0.7, A_s=2.1e-9, n_s=0.96,
        m_nu=[0.02, 0.1, 0.05], mass_split='list')

    with tempfile.TemporaryFile() as fp:
        pickle.dump(cosmo, fp)

        fp.seek(0)
        cosmo2 = pickle.load(fp)

    assert np.allclose(ccl.comoving_radial_distance(cosmo, 0.5),
                       ccl.comoving_radial_distance(cosmo2, 0.5),
                       atol=0, rtol=0)


def test_cosmology_lcdm():
    """Check that the default vanilla cosmology behaves
    as expected"""
    c1 = ccl.Cosmology(Omega_c=0.25,
                       Omega_b=0.05,
                       h=0.67, n_s=0.96,
                       sigma8=0.81)
    c2 = ccl.CosmologyVanillaLCDM()
    assert np.allclose(ccl.comoving_radial_distance(c1, 0.5),
                       ccl.comoving_radial_distance(c2, 0.5),
                       atol=0, rtol=0)


def test_cosmology_p18lcdm_raises():
    with pytest.raises(ValueError):
        kw = {'Omega_c': 0.1}
        ccl.CosmologyVanillaLCDM(**kw)


def test_cosmology_context():
    """Check that using a Cosmology object in a context manager
    frees C resources properly."""
    with ccl.Cosmology(
            Omega_c=0.25, Omega_b=0.05, h=0.7, A_s=2.1e-9, n_s=0.96,
            m_nu=np.array([0.02, 0.1, 0.05]), mass_split='list') as cosmo:
        # make sure it works
        assert not cosmo.has_distances
        ccl.comoving_radial_distance(cosmo, 0.5)
        assert cosmo.has_distances

    # make sure it does not!
    assert not hasattr(cosmo, "cosmo")
    assert not hasattr(cosmo, "_params")

    with pytest.raises(AttributeError):
        cosmo.has_growth


def test_pyccl_default_params():
    """Check that the Python-layer for setting the gsl and spline parameters
    works on par with the C-layer.
    """
    EPS_SCALEFAC_GROWTH = ccl.gsl_params["EPS_SCALEFAC_GROWTH"]

    # we will test with this parameter
    assert EPS_SCALEFAC_GROWTH == 1e-6

    # can be accessed as an attribute and as a dictionary item
    assert ccl.gsl_params.EPS_SCALEFAC_GROWTH == \
        ccl.gsl_params["EPS_SCALEFAC_GROWTH"]

    # can be assigned as an attribute
    ccl.gsl_params.EPS_SCALEFAC_GROWTH = 1e-5
    assert ccl.gsl_params["EPS_SCALEFAC_GROWTH"] == 1e-5  # cross-check

    ccl.gsl_params["EPS_SCALEFAC_GROWTH"] = 2e-6
    assert ccl.gsl_params.EPS_SCALEFAC_GROWTH == 2e-6

    # does not accept extra assignment
    with pytest.raises(KeyError):
        ccl.gsl_params.test = "hello_world"
    with pytest.raises(KeyError):
        ccl.gsl_params["test"] = "hello_world"

    # complains when we try to set A_SPLINE_MAX != 1.0
    ccl.spline_params.A_SPLINE_MAX = 1.0
    with pytest.raises(ValueError):
        ccl.spline_params.A_SPLINE_MAX = 0.9

    # complains when we try to change the spline type
    ccl.spline_params.A_SPLINE_TYPE = None
    with pytest.raises(TypeError):
        ccl.spline_params.A_SPLINE_TYPE = "something_else"

    # complains when we try to change the physical constants
    with pytest.raises(AttributeError):
        ccl.physical_constants.CLIGHT = 1

    # but if we unfreeze them, we can change them
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        ccl.physical_constants.unfreeze()
        ccl.physical_constants.CLIGHT = 1
    assert ccl.physical_constants.CLIGHT == 1
    ccl.physical_constants.freeze()
    ccl.physical_constants.reload()

    # verify that this has changed
    assert ccl.gsl_params.EPS_SCALEFAC_GROWTH != EPS_SCALEFAC_GROWTH

    # but now we reload it, so it should be the default again
    ccl.gsl_params.reload()
    assert ccl.gsl_params.EPS_SCALEFAC_GROWTH == EPS_SCALEFAC_GROWTH


def test_cosmology_default_params():
    """Check that the default params within Cosmology work as intended."""
    cosmo1 = ccl.CosmologyVanillaLCDM()
    v1 = cosmo1.cosmo.gsl_params.EPS_SCALEFAC_GROWTH

    ccl.gsl_params.EPS_SCALEFAC_GROWTH = v1*10
    cosmo2 = ccl.CosmologyVanillaLCDM()
    v2 = cosmo2.cosmo.gsl_params.EPS_SCALEFAC_GROWTH
    assert v2 == v1*10
    assert v2 != v1

    ccl.gsl_params.reload()
    cosmo3 = ccl.CosmologyVanillaLCDM()
    v3 = cosmo3.cosmo.gsl_params.EPS_SCALEFAC_GROWTH
    assert v3 == v1


def test_ccl_physical_constants_smoke():
    assert ccl.physical_constants.CLIGHT == ccl.ccllib.cvar.constants.CLIGHT


def test_ccl_global_parameters_repr():
    ccl.spline_params.reload()
    assert eval(repr(ccl.spline_params)) == ccl.spline_params._bak


def test_camb_sigma8_input():
    sigma8 = 0.85
    cosmo = ccl.Cosmology(
        Omega_c=0.25, Omega_b=0.05, h=0.67, n_s=0.96, sigma8=sigma8,
        transfer_function="boltzmann_camb"
    )
    assert np.isclose(cosmo.sigma8(), sigma8)



__all__ = ("MuSigmaMG",)

from dataclasses import dataclass

from . import ModifiedGravity


@dataclass
class MuSigmaMG(ModifiedGravity):
    """The :math:`\\mu`-:math:`\\Sigma` parametrization of modified gravity.

    Args:
        mu_0 (:obj:`float`): One of the parameters of the
            :math:`\\mu`-:math:`\\Sigma`
            modified gravity model. Defaults to 0.0
        sigma_0 (:obj:`float`): One of the parameters of the
            :math:`\\mu`-:math:`\\Sigma`
            modified gravity model. Defaults to 0.0
        c1_mg (:obj:`float`): MG parameter that enters in the scale
            dependence of :math:`\\mu`, affecting its large scale behavior.
            Default to 1. See, e.g., Eqs. (46) in
            `Ade et al. 2015 <https://arxiv.org/abs/1502.01590>`_,
            where their :math:`f_1` and :math:`f_2` functions are set equal
            to the commonly used ratio of dark energy density parameter at
            scale factor a over the dark energy density parameter today
        c2_mg (:obj:`float`): MG parameter that enters in the scale
            dependence of :math:`\\Sigma` affecting its large scale behavior.
            Default 1. See, e.g., Eqs. (47) in
            `Ade et al. 2015 <https://arxiv.org/abs/1502.01590>`_,
            where their :math:`f_1` and :math:`f_2` functions are set equal
            to the commonly used ratio of dark energy density parameter at
            scale factor a over the dark energy density parameter today
        lambda_mg (:obj:`float`): MG parameter that sets the start
            of dependence on :math:`c_1` and :math:`c_2` MG parameters.
            Defaults to 0.0 See, e.g., Eqs. (46) & (47) in
            `Ade et al. 2015 <https://arxiv.org/abs/1502.01590>`_,
            where their :math:`f_1` and :math:`f_2` functions are set equal
            to the commonly used ratio of dark energy density parameter at
            scale factor a over the dark energy density parameter today
    """

    parametrization: str = "mu_Sigma"
    mu_0: float = 0
    sigma_0: float = 0
    c1_mg: float = 1
    c2_mg: float = 1
    lambda_mg: float = 0



__all__ = ("ObjectLock",
           "UnlockInstance", "unlock_instance",
           "CustomRepr", "CustomEq",
           "CCLObject", "CCLAutoRepr", "CCLNamedClass",)

import functools
from abc import ABC, abstractmethod
from inspect import signature
from _thread import RLock

import numpy as np


class ObjectLock:
    """Control the lock state (immutability) of a ``CCLObject``."""
    _locked: bool = False
    _lock_id: int = None

    def __repr__(self):
        return f"{self.__class__.__name__}(locked={self.locked})"

    @property
    def locked(self):
        """Check if the object is locked."""
        return self._locked

    @property
    def active(self):
        """Check if an unlocking context manager is active."""
        return self._lock_id is not None

    def lock(self):
        """Lock the object."""
        self._locked = True
        self._lock_id = None

    def unlock(self, manager_id=None):
        """Unlock the object."""
        self._locked = False
        if manager_id is not None:
            self._lock_id = manager_id


class UnlockInstance:
    """Context manager that temporarily unlocks an immutable instance
    of ``CCLObject``.

    Parameters:
        instance (``CCLObject``):
            Instance of ``CCLObject`` to unlock within the scope
            of the context manager.
        mutate (``bool``):
            If the enclosed function mutates the object, the stored
            representation is automatically deleted.
    """

    def __init__(self, instance, *, mutate=True):
        self.instance = instance
        self.mutate = mutate
        # Define these attributes for easy access.
        self.id = id(self)
        self.thread_lock = RLock()
        # We want to catch and exit if the instance is not a CCLObject.
        # Hopefully this will be caught downstream.
        self.check_instance = isinstance(instance, CCLObject)
        if self.check_instance:
            self.object_lock = instance._object_lock

    def __enter__(self):
        if not self.check_instance:
            return

        with self.thread_lock:
            # Prevent simultaneous enclosing of a single instance.
            if self.object_lock.active:
                # Context manager already active.
                return

            # Unlock and store the fingerprint of this context manager so that
            # only this context manager is allowed to run on the instance.
            self.object_lock.unlock(manager_id=self.id)

    def __exit__(self, type, value, traceback):
        if not self.check_instance:
            return

        # If another context manager is running,
        # do nothing; otherwise reset.
        if self.id != self.object_lock._lock_id:
            return

        # Lock the instance on exit.
        # self.object_lock.lock()  # TODO: Uncomment for CCLv3.

    @classmethod
    def unlock_instance(cls, func=None, *, name=None, mutate=True):
        """Decorator that temporarily unlocks an instance of CCLObject.

        Arguments:
            func (``function``):
                Function which changes one of its ``CCLObject`` arguments.
            name (``str``):
                Name of the parameter to unlock. Defaults to the first one.
                If not a ``CCLObject`` the decorator will do nothing.
            mutate (``bool``):
                If after the function ``instance_old != instance_new``, the
                instance is mutated. If ``True``, the representation of the
                object will be reset.
        """
        if func is None:
            # called with parentheses
            return functools.partial(cls.unlock_instance, name=name,
                                     mutate=mutate)

        if not hasattr(func, "__signature__"):
            # store the function signature
            func.__signature__ = signature(func)
        names = list(func.__signature__.parameters.keys())
        name = names[0] if name is None else name  # default name
        if name not in names:
            # ensure the name makes sense
            raise NameError(f"{name} does not exist in {func.__name__}.")

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            bound = func.__signature__.bind(*args, **kwargs)
            with UnlockInstance(bound.arguments[name], mutate=mutate):
                return func(*args, **kwargs)
        return wrapper

    @classmethod
    def Funlock(cls, cl, name, mutate: bool):
        """Allow an instance to change or mutate when `name` is called."""
        func = vars(cl).get(name)
        if func is not None:
            newfunc = cls.unlock_instance(mutate=mutate)(func)
            setattr(cl, name, newfunc)


unlock_instance = UnlockInstance.unlock_instance


def is_equal(this, other):
    """Powerful helper for equivalence checking."""
    try:
        np.testing.assert_equal(this, other)
        return True
    except AssertionError:
        return False


class _CustomMethod:
    """Subclasses specifying a method string control whether the custom
    method is used in subclasses of ``CCLObject``.
    """

    def __init_subclass__(cls, *, method):
        super().__init_subclass__()
        if method not in vars(cls):
            raise ValueError(
                f"Subclass must contain a default {method} implementation.")
        cls._method = method
        cls._enabled: bool = True
        cls._classes: dict = {}

    @classmethod
    def register(cls, cl):
        """Register class to the dictionary of classes with custom methods."""
        if cls._method in vars(cls):
            cls._classes[cl] = getattr(cl, cls._method)

    @classmethod
    def enable(cls):
        """Enable the custom methods if they exist."""
        for cl, method in cls._classes.items():
            setattr(cl, cls._method, method)
        cls._enabled = True

    @classmethod
    def disable(cls):
        """Disable custom methods and fall back to Python defaults."""
        for cl in cls._classes.keys():
            default = getattr(cls, cls._method)
            setattr(cl, cls._method, default)
        cls._enabled = False


class CustomEq(_CustomMethod, method="__eq__"):
    """Controls the usage of custom ``__eq__`` for ``CCLObjects``."""

    def __eq__(self, other):
        # Default `eq`.
        return self is other


class CustomRepr(_CustomMethod, method="__repr__"):
    """Controls the usage of custom ``__repr__`` for ``CCLObjects``."""

    def __repr__(self):
        # Default `repr`.
        return object.__repr__(self)


class CCLObject(ABC):
    """Base for CCL objects.

    All CCL objects inherit ``__eq__`` and ``__hash__`` methods from here.
    Both methods rely on ``__repr__`` uniqueness. This aims to homogenize
    equivalence checking, and to standardize the use of hash.

    Overview
    --------
    ``CCLObjects`` inherit ``__hash__``, which consistently hashes the
    representation string. They also inherit ``__eq__`` which checks for
    representation equivalence.

    In the implemented scheme, each ``CCLObject`` may have its own, specialized
    ``__repr__`` method overloaded. Object representations have to be unique
    for equivalent objects. If no ``__repr__`` is provided, the default from
    ``object`` is used.

    Mutation
    --------
    ``CCLObjects`` are by default immutable. This aims to provide a failsafe
    mechanism, where, changing attributes has to trigger a re-computation
    of something else inside of the instance, rather than simply doing a value
    change.

    This immutability mechanism can be safely bypassed if a subclass defines an
    ``update_parameters`` method. ``CCLObjects`` temporarily unlock whenever
    this method is called.

    Internal State vs. Mutation
    ---------------------------
    Other methods that use ``setattr`` can only do that if they are decorated
    with ``@unlock_instance`` or if the particular code block that makes the
    change is enclosed within the ``UnlockInstance`` context manager.
    If neither is provided, an exception is raised.

    If such methods only change the instance's internal state, the decorator
    may be called with ``@unlock_instance(mutate=False)`` (or equivalently
    for the context manager ``UnlockInstance(..., mutate=False)``). Otherwise,
    the instance is assumed to have mutated.
    """

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        # 1. Store the initialization signature on import.
        cls.__signature__ = signature(cls.__init__)

        # 2. Register subclasses with custom dunder method implementations.
        CustomEq.register(cls)
        CustomRepr.register(cls)

        # 3. Unlock instance on specific methods.  # TODO: Uncomment for CCLv3.
        # UnlockInstance.Funlock(cls, "__init__", mutate=False)
        # UnlockInstance.Funlock(cls, "update_parameters", mutate=True)

    def __new__(cls, *args, **kwargs):
        # Populate every instance with an `ObjectLock` as attribute.
        instance = super().__new__(cls)
        object.__setattr__(instance, "_object_lock", ObjectLock())
        return instance

    def __setattr__(self, name, value):
        if self._object_lock.locked:
            raise AttributeError("CCL objects can only be updated via "
                                 "`update_parameters`, if implemented.")
        object.__setattr__(self, name, value)

    def update_parameters(self, **kwargs):
        name = self.__class__.__name__
        raise NotImplementedError(f"{name} objects are immutable.")

    def __repr__(self):
        # By default we use `__repr__` from `object`.
        return object.__repr__(self)

    def __hash__(self):
        # `__hash__` makes use of the `repr` of the object,
        # so we have to make sure that the `repr` is unique.
        return hash(repr(self))

    def __eq__(self, other):
        # Exit early if it is the same object.
        if self is other:
            return True
        # Two same-type objects are equal if their representations are equal.
        if type(self) is not type(other):
            return False
        # Compare the attributes listed in `__eq_attrs__`.
        if hasattr(self, "__eq_attrs__"):
            for attr in self.__eq_attrs__:
                if not is_equal(getattr(self, attr), getattr(other, attr)):
                    return False
            return True
        return False


class CCLAutoRepr(CCLObject):
    """Base for objects with automatic representation. Representations
    for instances are built from a list of attribute names specified as
    a class variable in ``__repr_attrs__`` (acting as a hook).

    Example:
        The representation (also hash) of instances of the following class
        is built based only on the attributes specified in ``__repr_attrs__``:

        >>> class MyClass(CCLAutoRepr):
            __repr_attrs__ = ("a", "b", "other")
            def __init__(self, a=1, b=2, c=3, d=4, e=5):
                self.a = a
                self.b = b
                self.c = c
                self.other = d + e

        >>> repr(MyClass(6, 7, 8, 9, 10))
            <__main__.MyClass>
                a = 6
                b = 7
                other = 19
    """

    def __repr__(self):
        # Build string from specified `__repr_attrs__` or use Python's default.
        # Subclasses overriding `__repr__`, stop using `__repr_attrs__`.
        if hasattr(self.__class__, "__repr_attrs__"):
            from .repr_ import build_string_from_attrs
            return build_string_from_attrs(self)
        return object.__repr__(self)


class CCLNamedClass(CCLObject):
    """Base for objects that contain methods ``from_name()`` and
    ``create_instance()``.

    Implementation
    --------------
    Subclasses must define a ``name`` class attribute which allows the tree to
    be searched to retrieve the particular model, using its name.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """Class attribute denoting the name of the model."""

    @classmethod
    def _subclasses(cls):
        # This helper returns a set of all subclasses.
        direct_subs = cls.__subclasses__()
        deep_subs = [sub for cl in direct_subs for sub in cl._subclasses()]
        return set(direct_subs).union(deep_subs)

    @classmethod
    def from_name(cls, name):
        """Obtain particular model."""
        mod = {p.name: p for p in cls._subclasses() if hasattr(p, "name")}
        if name not in mod:
            raise KeyError(f"Invalid model {name}.")
        return mod[name]

    @classmethod
    def create_instance(cls, input_, **kwargs):
        """Process the input and generate an object of the class.
        Input can be an instance of the class, or a name string.
        Optional ``**kwargs`` may be passed.
        """
        if isinstance(input_, cls):
            return input_
        if isinstance(input_, str):
            class_ = cls.from_name(input_)
            return class_(**kwargs)
        good, bad = cls.__name__, type(input_).__name__
        raise TypeError(f"Expected {good} or str but received {bad}.")



__all__ = ("linear_power", "nonlin_power", "linear_matter_power",
           "nonlin_matter_power", "sigmaM", "sigmaR", "sigmaV", "sigma8",
           "kNL",)

import numpy as np

from . import DEFAULT_POWER_SPECTRUM, check, lib


def linear_power(cosmo, k, a, *, p_of_k_a=DEFAULT_POWER_SPECTRUM):
    """The linear power spectrum.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological parameters.
        k (:obj:`float` or `array`): Wavenumber; :math:`{\\rm Mpc}^{-1}`.
        a (:obj:`float` or `array`): Scale factor.
        p_of_k_a (:obj:`str`): string specifying the power spectrum to
            compute (which should be stored in ``cosmo``). Defaults to
            the linear matter power spectrum.

    Returns:
        (:obj:`float` or `array`): Linear power spectrum.
    """
    return cosmo.get_linear_power(p_of_k_a)(k, a, cosmo)


def nonlin_power(cosmo, k, a, *, p_of_k_a=DEFAULT_POWER_SPECTRUM):
    """The non-linear power spectrum.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological parameters.
        k (:obj:`float` or `array`): Wavenumber; :math:`{\\rm Mpc}^{-1}`.
        a (:obj:`float` or `array`): Scale factor.
        p_of_k_a (:obj:`str`): string specifying the power spectrum to
            compute (which should be stored in ``cosmo``). Defaults to
            the non-linear matter power spectrum.

    Returns:
        (:obj:`float` or `array`): Non-linear power spectrum.
    """
    return cosmo.get_nonlin_power(p_of_k_a)(k, a, cosmo)


def linear_matter_power(cosmo, k, a):
    """The linear matter power spectrum

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological parameters.
        k (:obj:`float` or `array`): Wavenumber; :math:`{\\rm Mpc}^{-1}`.
        a (:obj:`float` or `array`): Scale factor.

    Returns:
        (:obj:`float` or `array`): Linear matter power spectrum;
        :math:`{\\rm Mpc}^3`.
    """
    return cosmo.linear_power(k, a, p_of_k_a=DEFAULT_POWER_SPECTRUM)


def nonlin_matter_power(cosmo, k, a):
    """The nonlinear matter power spectrum

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological parameters.
        k (:obj:`float` or `array`): Wavenumber; :math:`{\\rm Mpc}^{-1}`.
        a (:obj:`float` or `array`): Scale factor.

    Returns:
        (:obj:`float` or `array`): Nonlinear matter power spectrum;
        :math:`{\\rm Mpc}^3`.
    """
    return cosmo.nonlin_power(k, a, p_of_k_a=DEFAULT_POWER_SPECTRUM)


def sigmaM(cosmo, M, a):
    """RMS on the scale of a halo of mass :math:`M`. Calculated
    as :math:`\\sigma_R` (see :func:`sigmaR`) with :math:`R` being
    the Lagrangian radius of a halo of mass :math:`M` (see
    :func:`~pyccl.halos.massdef.mass2radius_lagrangian`).

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological parameters.
        M (:obj:`float` or `array`): Halo masses.
        a (:obj:`float`): scale factor.

    Returns:
        (:obj:`float` or `array`): RMS variance of halo mass.
    """
    cosmo.compute_sigma()

    logM = np.log10(np.atleast_1d(M))
    status = 0
    sigM, status = lib.sigM_vec(cosmo.cosmo, a, logM,
                                len(logM), status)
    check(status, cosmo=cosmo)
    if np.ndim(M) == 0:
        sigM = sigM[0]
    return sigM


def sigmaR(cosmo, R, a=1, *, p_of_k_a=DEFAULT_POWER_SPECTRUM):
    """RMS of the matter overdensity a top-hat sphere of radius :math:`R`.

    .. math::
        \\sigma_R^2(z)=\\frac{1}{2\\pi^2}\\int dk\\,k^2\\,P(k,z)\\,
        |W(kR)|^2,

    with :math:`W(x)=(3\\sin(x)-x\\cos(x))/x^3`.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological parameters.
        R (:obj:`float` or `array`): Radius; Mpc.
        a (:obj:`float`): optional scale factor.
        p_of_k_a (:class:`~pyccl.pk2d.Pk2D`, :obj:`str` or :obj:`None`):
            power spectrum to integrate. If a string, it must correspond to
            one of the linear power spectra stored in ``cosmo`` (e.g.
            ``'delta_matter:delta_matter'``).

    Returns:
        (:obj:`float` or `array`): :math:`\\sigma_R`.
    """
    psp = cosmo.parse_pk2d(p_of_k_a, is_linear=True)
    status = 0
    R_use = np.atleast_1d(R)
    sR, status = lib.sigmaR_vec(cosmo.cosmo, psp, a, R_use, R_use.size, status)
    check(status, cosmo)
    if np.ndim(R) == 0:
        sR = sR[0]
    return sR


def sigmaV(cosmo, R, a=1, *, p_of_k_a=DEFAULT_POWER_SPECTRUM):
    """RMS of the linear displacement field in a top-hat sphere of radius R.

    .. math::
        \\sigma_V^2(z)=\\frac{1}{6\\pi^2}\\int dk\\,P(k,z)\\,|W(kR)|^2,

    with :math:`W(x)=(3\\sin(x)-x\\cos(x))/x^3`.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological parameters.
        R (:obj:`float` or `array`): Radius; Mpc.
        a (:obj:`float`): optional scale factor.
        p_of_k_a (:class:`~pyccl.pk2d.Pk2D`, :obj:`str` or :obj:`None`):
            power spectrum to integrate. If a string, it must correspond to
            one of the linear power spectra stored in ``cosmo`` (e.g.
            ``'delta_matter:delta_matter'``).

    Returns:
        (:obj:`float` or `array`): :math:`\\sigma_V` (:math:`{\\rm Mpc}`).
    """
    psp = cosmo.parse_pk2d(p_of_k_a, is_linear=True)
    status = 0
    R_use = np.atleast_1d(R)
    sV, status = lib.sigmaV_vec(cosmo.cosmo, psp, a, R_use, R_use.size, status)
    check(status, cosmo)
    if np.ndim(R) == 0:
        sV = sV[0]
    return sV


def sigma8(cosmo, *, p_of_k_a=DEFAULT_POWER_SPECTRUM):
    """RMS variance in a top-hat sphere of radius :math:`8\\,{\\rm Mpc}/h`,
    (with the value of :math:`h` extracted from ``cosmo``) at :math:`z=0`.


    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological parameters.
        p_of_k_a (:class:`~pyccl.pk2d.Pk2D`, :obj:`str` or :obj:`None`):
            power spectrum  to integrate. If a string, it must correspond to
            one of the linear power spectra stored in ``cosmo`` (e.g.
            ``'delta_matter:delta_matter'``).

    Returns:
        :obj:`float`: :math:`\\sigma_8`.
    """
    sig8 = cosmo.sigmaR(8/cosmo["h"], p_of_k_a=p_of_k_a)
    if np.isnan(cosmo["sigma8"]):
        cosmo._fill_params(sigma8=sig8)
    return sig8


def kNL(cosmo, a, *, p_of_k_a=DEFAULT_POWER_SPECTRUM):
    r"""Non-linear scale :math:`k_{\rm NL}`. Calculated based on Lagrangian
    perturbation theory as the inverse of the rms of the displacement
    field, i.e.:

    .. math::
        k_{\rm NL}(z) = \left[\frac{1}{6\pi^2} \int dk\,P_L(k,z)\right]^{-1/2}.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological parameters.
        a (:obj:`float` or `array`): Scale factor(s), normalized to 1 today.
        p_of_k_a (:class:`~pyccl.pk2d.Pk2D`, :obj:`str` or :obj:`None`):
            power spectrum  to integrate. If a string, it must correspond to
            one of the linear power spectra stored in ``cosmo`` (e.g.
            ``'delta_matter:delta_matter'``).

    Returns:
        :obj:`float` or `array`: :math:`k_{\rm NL}`.
    """
    psp = cosmo.parse_pk2d(p_of_k_a, is_linear=True)
    status = 0
    a_use = np.atleast_1d(a)
    knl, status = lib.kNL_vec(cosmo.cosmo, psp, a_use, a_use.size, status)
    check(status, cosmo)
    if np.ndim(a) == 0:
        knl = knl[0]
    return knl



import numpy as np
import pytest
import pyccl as ccl


def test_nonlin_camb_power():
    import camb

    logT_AGN = 7.93
    Omega_c = 0.25
    Omega_b = 0.05
    A_s = 2.1e-9
    n_s = 0.97
    h = 0.7
    # Needs to be set for good agreements between CCL and CAMB
    T_CMB = 2.7255

    p = camb.CAMBparams(WantTransfer=True,
                        NonLinearModel=camb.nonlinear.Halofit(
                            halofit_version="mead2020_feedback",
                            HMCode_logT_AGN=logT_AGN))
    # This affects k_min
    p.WantCls = False
    p.DoLensing = False
    p.Want_CMB = False
    p.Want_CMB_lensing = False
    p.Want_cl_2D_array = False
    p.set_cosmology(H0=h*100, omch2=Omega_c*h**2, ombh2=Omega_b*h**2,
                    mnu=0.0, TCMB=T_CMB)
    p.share_delta_neff = False
    p.InitPower.set_params(As=A_s, ns=n_s)

    z = [0.0, 0.5, 1.0, 1.5]
    p.set_matter_power(redshifts=z, kmax=10.0, nonlinear=True)
    p.set_for_lmax(5000)

    r = camb.get_results(p)

    k, z, pk_nonlin_camb = r.get_nonlinear_matter_power_spectrum(
        hubble_units=False, k_hunit=False)

    ccl_cosmo = ccl.Cosmology(
        Omega_c=Omega_c, Omega_b=Omega_b, h=h, m_nu=0.0,
        A_s=A_s, n_s=n_s,
        transfer_function="boltzmann_camb",
        matter_power_spectrum="camb",
        extra_parameters={"camb": {"halofit_version": "mead2020_feedback",
                                   "HMCode_logT_AGN": logT_AGN}})

    for z_, pk_camb in zip(z, pk_nonlin_camb):
        pk_nonlin_ccl = ccl.nonlin_matter_power(ccl_cosmo, k, 1/(1+z_))

        assert np.allclose(pk_camb, pk_nonlin_ccl, rtol=3e-5)


def test_nonlin_camb_power_raises():
    # Test that it raises when (trf, mps) == (no camb, camb).
    with pytest.raises(ccl.CCLError):
        ccl.CosmologyVanillaLCDM(transfer_function="boltzmann_class",
                                 matter_power_spectrum="camb")



import numpy as np
import pyccl as ccl


def test_szcl():
    fsky = 1.
    COSMO = ccl.Cosmology(
        Omega_b=0.05,
        Omega_c=0.25,
        h=0.7,
        n_s=0.9645,
        A_s=2.02E-9,
        Neff=3.046, T_CMB=2.725,
        transfer_function='boltzmann_class')
    bm = np.loadtxt("benchmarks/data/sz_cl_P13_szpowerspectrum.txt",
                    unpack=True)
    l_bm = bm[0]
    cl_bm = bm[1]
    tll_bm = np.loadtxt("benchmarks/data/tSZ_trispectrum_ref_for_cobaya.txt")
    fac = 2*np.pi/(l_bm*(l_bm+1)*1E12)

    cl_bm *= fac
    tll_bm *= fac[:, None]*fac[None, :]/(4*np.pi*fsky)
    mass_def = ccl.halos.MassDef(500, 'critical')
    hmf = ccl.halos.MassFuncTinker08(mass_def=mass_def)
    hbf = ccl.halos.HaloBiasTinker10(mass_def=mass_def)
    hmc = ccl.halos.HMCalculator(mass_function=hmf, halo_bias=hbf,
                                 mass_def=mass_def)
    prf = ccl.halos.HaloProfilePressureGNFW(mass_def=mass_def)
    prf.update_parameters(mass_bias=1./1.41, x_out=6.)
    tr = ccl.tSZTracer(COSMO, z_max=3.)

    # Power spectrum
    pk = ccl.halos.halomod_Pk2D(COSMO, hmc, prf, get_2h=False)
    cl = ccl.angular_cl(COSMO, tr, tr, ell=l_bm, p_of_k_a=pk)

    # Covariance
    lk_arr = np.log(np.geomspace(1E-4, 1E2, 256))
    a_arr = 1./(1+np.linspace(0, 3., 20))[::-1]
    tkk = ccl.halos.halomod_Tk3D_1h(COSMO, hmc, prf,
                                    lk_arr=lk_arr, a_arr=a_arr,
                                    use_log=True)
    tll = ccl.angular_cl_cov_cNG(COSMO, tr, tr, ell=l_bm, t_of_kk_a=tkk,
                                 fsky=fsky)

    assert np.all(np.fabs(cl/cl_bm-1) < 2E-2)
    assert np.all(np.fabs(tll/tll_bm-1) < 5E-2)



import numpy as np
import pyccl as ccl
import time


def test_timing():
    ls = np.unique(np.geomspace(2, 2000, 128).astype(int)).astype(float)
    nl = len(ls)
    b_g = np.array([1.376695, 1.451179, 1.528404,
                    1.607983, 1.689579, 1.772899,
                    1.857700, 1.943754, 2.030887,
                    2.118943])
    dNdz_file = np.load('benchmarks/data/dNdzs.npz')
    z_s = dNdz_file['z_sh']
    dNdz_s = dNdz_file['dNdz_sh'].T
    z_g = dNdz_file['z_cl']
    dNdz_g = dNdz_file['dNdz_cl'].T
    nt = len(dNdz_s) + len(dNdz_g)
    nx = (nt*(nt+1))//2

    cosmo = ccl.CosmologyVanillaLCDM(transfer_function='boltzmann_class')
    cosmo.compute_nonlin_power()

    start = time.time()
    t_g = [ccl.NumberCountsTracer(cosmo, has_rsd=True, dndz=(z_g, ng),
                                  bias=(z_g, np.full(len(z_g), b)))
           for ng, b in zip(dNdz_g, b_g)]
    t_s = [ccl.WeakLensingTracer(cosmo, dndz=(z_s, ns)) for ns in dNdz_s]
    t_all = t_g + t_s

    cls = np.zeros([nx, nl])
    ind1, ind2 = np.triu_indices(nt)
    for ix, (i1, i2) in enumerate(zip(ind1, ind2)):
        cls[ix, :] = ccl.angular_cl(cosmo, t_all[i1], t_all[i2], ls)
    end = time.time()
    t_seconds = end - start
    print(end-start)
    assert t_seconds < 3.



import numpy as np
import pytest
import pyccl as ccl
from .test_cclobject import check_eq_repr_hash


def test_Tk3D_eq_repr_hash():
    # Test eq, repr, hash for Tk3D.
    cosmo = ccl.CosmologyVanillaLCDM(transfer_function="bbks")
    cosmo.compute_linear_power()
    PK1 = cosmo.get_linear_power()

    # 1. Using a factorizable Tk3D object.
    a_arr, lk_arr, pk_arr = PK1.get_spline_arrays()
    TK1 = ccl.Tk3D(a_arr=a_arr, lk_arr=lk_arr,
                   pk1_arr=pk_arr, pk2_arr=pk_arr, is_logt=False)
    TK2 = ccl.Tk3D(a_arr=a_arr, lk_arr=lk_arr,
                   pk1_arr=pk_arr, pk2_arr=pk_arr, is_logt=False)
    assert check_eq_repr_hash(TK1, TK2)

    TK3 = ccl.Tk3D(a_arr=a_arr, lk_arr=lk_arr,
                   pk1_arr=2*pk_arr, pk2_arr=2*pk_arr, is_logt=False)
    assert check_eq_repr_hash(TK1, TK3, equal=False)

    # 2. Using a non-factorizable Tk3D object.
    a_arr_2 = np.arange(0.5, 0.9, 0.1)
    lk_arr_2 = np.linspace(-2, 1, 8)
    TK4 = ccl.Tk3D(
        a_arr=a_arr_2, lk_arr=lk_arr_2,
        tkk_arr=np.ones((a_arr_2.size, lk_arr_2.size, lk_arr_2.size)))
    TK5 = ccl.Tk3D(
        a_arr=a_arr_2, lk_arr=lk_arr_2,
        tkk_arr=np.ones((a_arr_2.size, lk_arr_2.size, lk_arr_2.size)))
    assert check_eq_repr_hash(TK4, TK5)

    TK6 = ccl.Tk3D(
        a_arr=a_arr_2, lk_arr=lk_arr_2,
        tkk_arr=2*np.ones((a_arr_2.size, lk_arr_2.size, lk_arr_2.size)))
    assert check_eq_repr_hash(TK4, TK6, equal=False)

    # edge-case: comparing different types
    assert check_eq_repr_hash(TK1, 1, equal=False)

    # edge-case: empty objects
    tka1, tka2 = [ccl.Tk3D.__new__(ccl.Tk3D) for _ in range(2)]
    assert check_eq_repr_hash(tka1, tka2)

    # edge-case: only one Tk is factorizable (exits early)
    assert check_eq_repr_hash(TK1, TK4, equal=False)

    # edge-case: different extrapolation orders
    a_arr, lk_arr, pk_arr = PK1.get_spline_arrays()
    t1 = ccl.Tk3D(a_arr=a_arr, lk_arr=lk_arr, pk1_arr=pk_arr, pk2_arr=pk_arr,
                  is_logt=False, extrap_order_lok=0)
    t2 = ccl.Tk3D(a_arr=a_arr, lk_arr=lk_arr, pk1_arr=pk_arr, pk2_arr=pk_arr,
                  is_logt=False, extrap_order_lok=1)
    assert check_eq_repr_hash(t1, t2, equal=False)


def kf(k):
    return (k/0.1)**(-1)


def af(a):
    return np.exp(a-1)


def fka1f(k, a):
    return kf(k)*af(a)**2


def fka2f(k, a):
    return (kf(k))**2.1*af(a)**2


def tkkaf(k1, k2, a):
    return fka1f(k1, a)*fka2f(k2, a)


def get_arrays(islog=True):
    a_arr = np.linspace(0.05, 1., 10)
    k_arr = np.geomspace(1E-4, 1E2, 10)
    lk_arr = np.log(k_arr)
    fka1_arr = np.array([fka1f(k_arr, a) for a in a_arr])
    fka2_arr = np.array([fka2f(k_arr, a) for a in a_arr])
    tkka_arr = np.array([tkkaf(k_arr[None, :],
                               k_arr[:, None], a)
                         for a in a_arr])
    if islog:
        fka1_arr = np.log(fka1_arr)
        fka2_arr = np.log(fka2_arr)
        tkka_arr = np.log(tkka_arr)
    return (a_arr, lk_arr, fka1_arr,
            fka2_arr, tkka_arr)


def test_tk3d_errors():
    """
    Test initialization of Pk2D objects
    """

    (a_arr, lk_arr, fka1_arr, fka2_arr, tkka_arr) = get_arrays()

    # Decreasing a
    with pytest.raises(ValueError):
        ccl.Tk3D(a_arr=a_arr[::-1], lk_arr=lk_arr,
                 tkk_arr=tkka_arr)
    # Decreasing lk
    with pytest.raises(ValueError):
        ccl.Tk3D(a_arr=a_arr, lk_arr=lk_arr[::-1],
                 tkk_arr=tkka_arr)
    # Non monotonic
    a2 = a_arr.copy()
    a2[1] = a2[0]
    with pytest.raises(ValueError):
        ccl.Tk3D(a_arr=a2, lk_arr=lk_arr, tkk_arr=tkka_arr)

    # If no input
    with pytest.raises(TypeError):
        ccl.Tk3D()

    # No input tkk or fkas
    with pytest.raises(ValueError):
        ccl.Tk3D(a_arr=a_arr, lk_arr=lk_arr)

    # Missing one fka factor
    with pytest.raises(ValueError):
        ccl.Tk3D(a_arr=a_arr, lk_arr=lk_arr, pk2_arr=fka2_arr)

    # fka has wrong shape
    with pytest.raises(ValueError):
        ccl.Tk3D(a_arr=a_arr, lk_arr=lk_arr, pk1_arr=tkka_arr)

    # tkka has wrong shape
    with pytest.raises(ValueError):
        ccl.Tk3D(a_arr=a_arr, lk_arr=lk_arr, tkk_arr=fka1_arr)

    # Wrong extrapolation orders
    with pytest.raises(ValueError):
        ccl.Tk3D(a_arr=a_arr, lk_arr=lk_arr,
                 tkk_arr=tkka_arr, extrap_order_hik=-1)
    with pytest.raises(ValueError):
        ccl.Tk3D(a_arr=a_arr, lk_arr=lk_arr,
                 tkk_arr=tkka_arr, extrap_order_lok=2)


def test_tk3d_smoke():
    """Make sure it works once."""
    (a_arr, lk_arr, fka1_arr, fka2_arr, tkka_arr) = get_arrays()
    tsp1 = ccl.Tk3D(a_arr=a_arr, lk_arr=lk_arr, pk1_arr=fka1_arr,
                    pk2_arr=fka2_arr)
    tsp2 = ccl.Tk3D(a_arr=a_arr, lk_arr=lk_arr, tkk_arr=tkka_arr)
    assert not np.isnan(tsp1(1E-2, 0.5))
    assert not np.isnan(tsp2(1E-2, 0.5))


def test_tk3d_delete():
    """Check that ccl.Tk3D.__del__ works."""
    (a_arr, lk_arr, fka1_arr, fka2_arr, tkka_arr) = get_arrays()
    tsp = ccl.Tk3D(a_arr=a_arr, lk_arr=lk_arr, pk1_arr=fka1_arr,
                   pk2_arr=fka2_arr)
    # This should not cause an ignored exception
    del tsp


@pytest.mark.parametrize('is_product', [True, False])
def test_tk3d_eval(is_product):
    (a_arr, lk_arr, fka1_arr, fka2_arr, tkka_arr) = get_arrays()
    if is_product:
        tsp = ccl.Tk3D(a_arr=a_arr, lk_arr=lk_arr, pk1_arr=fka1_arr,
                       pk2_arr=fka2_arr)
    else:
        tsp = ccl.Tk3D(a_arr=a_arr, lk_arr=lk_arr, tkk_arr=tkka_arr)

    # Test at single point
    ktest = 0.7
    atest = 0.5
    ptrue = tkkaf(ktest, ktest, atest)
    phere = tsp(ktest, atest)
    assert np.allclose(phere, ptrue, atol=0, rtol=1e-6)

    ktest = 5E-5
    atest = 0.5
    ptrue = tkkaf(ktest, ktest, atest)
    phere = tsp(ktest, atest)
    assert np.allclose(phere, ptrue, atol=0, rtol=1e-6)

    # Test at array of points
    ktest = np.logspace(-3, 1, 10)
    ptrue = tkkaf(ktest[None, :], ktest[:, None], atest)
    phere = tsp(ktest, atest)
    assert np.allclose(phere, ptrue, atol=0, rtol=1e-6)


def test_tk3d_call():
    # Test `__call__` and `__bool__`
    (a_arr, lk_arr, fka1_arr, fka2_arr, tkka_arr) = get_arrays()
    tsp = ccl.Tk3D(a_arr=a_arr, lk_arr=lk_arr, tkk_arr=tkka_arr)
    assert bool(tsp) is tsp.has_tsp
    assert np.allclose(np.array([tsp(np.exp(lk_arr), a) for a in a_arr]),
                       tsp(np.exp(lk_arr), a_arr), rtol=1e-15)


@pytest.mark.parametrize('is_product', [True, False])
def test_tk3d_spline_arrays(is_product):
    (a_arr, lk_arr, fka1_arr, fka2_arr, tkka_arr) = get_arrays()
    if is_product:
        tsp = ccl.Tk3D(a_arr=a_arr, lk_arr=lk_arr,
                       pk1_arr=fka1_arr, pk2_arr=fka2_arr)
    else:
        tsp = ccl.Tk3D(a_arr=a_arr, lk_arr=lk_arr, tkk_arr=tkka_arr)

    a_get, lk_get1, lk_get2, out = tsp.get_spline_arrays()
    assert np.allclose(a_get, a_arr, rtol=1e-15)
    assert np.allclose(lk_get1, lk_arr, rtol=1e-15)
    assert np.allclose(lk_get2, lk_arr, rtol=1e-15)

    if is_product:
        assert np.allclose(np.log(out[0]), fka1_arr, rtol=1e-15)
        assert np.allclose(np.log(out[1]), fka2_arr, rtol=1e-15)
    else:
        assert np.allclose(np.log(out[0]), tkka_arr, rtol=1e-15)


def test_tk3d_spline_arrays_raises():
    (a_arr, lk_arr, fka1_arr, fka2_arr, tkka_arr) = get_arrays()
    tsp = ccl.Tk3D(a_arr=a_arr, lk_arr=lk_arr, tkk_arr=tkka_arr)

    ccl.lib.f3d_t_free(tsp.tsp)
    delattr(tsp, "tsp")

    with pytest.raises(ValueError):
        tsp.get_spline_arrays()



__all__ = ("HMCalculator",)

import numpy as np

from .. import CCLAutoRepr, unlock_instance
from .. import physical_constants as const
from . import MassDef
from ..pyutils import _spline_integrate


class HMCalculator(CCLAutoRepr):
    """This class implements a set of methods that can be used to
    compute various halo model quantities. A lot of these quantities
    will involve integrals of the sort:

    .. math::
       \\int dM\\,n(M,a)\\,f(M,k,a),

    where :math:`n(M,a)` is the halo mass function, and :math:`f` is
    an arbitrary function of mass, scale factor and Fourier scales.

    Args:
        mass_function (str or :class:`~pyccl.halos.halo_model_base.MassFunc`):
            the mass function to use
        halo_bias (str or :class:`~pyccl.halos.halo_model_base.HaloBias`):
            the halo bias function to use
        mass_def (str or :class:`~pyccl.halos.massdef.MassDef`):
            the halo mass definition to use
        log10M_min (:obj:`float`): lower bound of the mass integration range
            (base-10 logarithmic).
        log10M_max (:obj:`float`): lower bound of the mass integration range
            (base-10 logarithmic).
        nM (:obj:`int`): number of uniformly-spaced samples in :math:`\\log_{10}(M)`
            to be used in the mass integrals.
        integration_method_M (:obj:`str`): integration method to use
            in the mass integrals. Options: "simpson" and "spline".
    """ # noqa
    __repr_attrs__ = __eq_attrs__ = (
        "mass_function", "halo_bias", "mass_def", "precision",)

    def __init__(self, *, mass_function, halo_bias, mass_def=None,
                 log10M_min=8., log10M_max=16., nM=128,
                 integration_method_M='simpson'):
        # Initialize halo model ingredients.
        out = MassDef.from_specs(mass_def, mass_function=mass_function,
                                 halo_bias=halo_bias)
        if len(out) != 3:
            raise ValueError("A valid mass function and halo bias is "
                             "needed")
        self.mass_def, self.mass_function, self.halo_bias = out

        self.precision = {
            'log10M_min': log10M_min, 'log10M_max': log10M_max, 'nM': nM,
            'integration_method_M': integration_method_M}
        self._lmass = np.linspace(log10M_min, log10M_max, nM)
        self._mass = 10.**self._lmass
        self._m0 = self._mass[0]

        if integration_method_M == "simpson":
            from scipy.integrate import simpson
            self._integrator = simpson
        elif integration_method_M == "spline":
            self._integrator = self._integ_spline
        else:
            raise ValueError("Invalid integration method.")

        # Cache last results for mass function and halo bias.
        self._cosmo_mf = self._cosmo_bf = None
        self._a_mf = self._a_bf = -1

    def _integ_spline(self, fM, log10M):
        # Spline integrator
        return _spline_integrate(log10M, fM, log10M[0], log10M[-1])

    def _check_mass_def(self, *others):
        # Verify that internal & external mass definitions are consistent.
        if set([x.mass_def for x in others]) != set([self.mass_def]):
            raise ValueError("Inconsistent mass definitions.")

    @unlock_instance(mutate=False)
    def _get_mass_function(self, cosmo, a, rho0):
        # Compute the mass function at this cosmo and a.
        if a != self._a_mf or cosmo != self._cosmo_mf:
            self._mf = self.mass_function(cosmo, self._mass, a)
            integ = self._integrator(self._mf*self._mass, self._lmass)
            self._mf0 = (rho0 - integ) / self._m0
            self._cosmo_mf, self._a_mf = cosmo, a  # cache

    @unlock_instance(mutate=False)
    def _get_halo_bias(self, cosmo, a, rho0):
        # Compute the halo bias at this cosmo and a.
        if a != self._a_bf or cosmo != self._cosmo_bf:
            self._bf = self.halo_bias(cosmo, self._mass, a)
            integ = self._integrator(self._mf*self._bf*self._mass, self._lmass)
            self._mbf0 = (rho0 - integ) / self._m0
            self._cosmo_bf, self._a_bf = cosmo, a  # cache

    def _get_ingredients(self, cosmo, a, *, get_bf):
        """Compute mass function and halo bias at some scale factor."""
        rho0 = const.RHO_CRITICAL * cosmo["Omega_m"] * cosmo["h"]**2
        self._get_mass_function(cosmo, a, rho0)
        if get_bf:
            self._get_halo_bias(cosmo, a, rho0)

    def _integrate_over_mf(self, array_2):
        #  ∫ dM n(M) f(M)
        i1 = self._integrator(self._mf * array_2, self._lmass)
        return i1 + self._mf0 * array_2[..., 0]

    def _integrate_over_mbf(self, array_2):
        #  ∫ dM n(M) b(M) f(M)
        i1 = self._integrator(self._mf * self._bf * array_2, self._lmass)
        return i1 + self._mbf0 * array_2[..., 0]

    def integrate_over_massfunc(self, func, cosmo, a):
        """ Returns the integral over mass of a given funcion times
        the mass function:

        .. math::
            \\int dM\\,n(M,a)\\,f(M)

        Args:
            func (:obj:`callable`): a function accepting an array of halo masses
                as a single argument, and returning an array of the
                same size.
            cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
            a (:obj:`float`): scale factor.

        Returns:
            :obj:`float`: integral value.
        """ # noqa
        fM = func(self._mass)
        self._get_ingredients(cosmo, a, get_bf=False)
        return self._integrate_over_mf(fM)

    def number_counts(self, cosmo, *, selection,
                      a_min=None, a_max=1.0, na=128):
        """ Solves the integral:

        .. math::
            nc(sel) = \\int dM\\int da\\,\\frac{dV}{dad\\Omega}\\,
            n(M,a)\\,sel(M,a)

        where :math:`n(M,a)` is the halo mass function, and
        :math:`sel(M,a)` is the selection function as a function of halo mass
        and scale factor.

        Note that the selection function is normalized to integrate to unity
        and assumed to represent the selection probaility per unit scale factor
        and per unit mass.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
            selection (:obj:`callable`): function of mass and scale factor
                that returns the selection function. This function
                should take in floats or arrays with a signature ``sel(m, a)``
                and return an array with shape ``(len(m), len(a))`` according
                to the numpy broadcasting rules.
            a_min (:obj:`float`): the minimum scale factor at which to start integrals
                over the selection function.
                Default: value of ``cosmo.cosmo.spline_params.A_SPLINE_MIN``
            a_max (:obj:`float`): the maximum scale factor at which to end integrals
                over the selection function.
            na (:obj:`int`): number of samples in scale factor to be used in
                the integrals.

        Returns:
            :obj:`float`: the total number of clusters/halos.
        """ # noqa
        # get a values for integral
        if a_min is None:
            a_min = cosmo.cosmo.spline_params.A_SPLINE_MIN
        a = np.linspace(a_min, a_max, na)

        # compute the volume element
        abs_dzda = 1 / a / a
        dc = cosmo.comoving_angular_distance(a)
        ez = cosmo.h_over_h0(a)
        dh = const.CLIGHT_HMPC / cosmo['h']
        dvdz = dh * dc**2 / ez
        dvda = dvdz * abs_dzda

        # now do m intergrals in a loop
        mint = np.zeros_like(a)
        for i, _a in enumerate(a):
            self._get_ingredients(cosmo, _a, get_bf=False)
            _selm = np.atleast_2d(selection(self._mass, _a)).T
            mint[i] = self._integrator(
                dvda[i] * self._mf[..., :] * _selm[..., :],
                self._lmass
            )

        # now do scale factor integral
        return self._integrator(mint, a)

    def I_0_1(self, cosmo, k, a, prof):
        """ Solves the integral:

        .. math::
            I^0_1(k,a|u) = \\int dM\\,n(M,a)\\,\\langle u(k,a|M)\\rangle,

        where :math:`n(M,a)` is the halo mass function, and
        :math:`\\langle u(k,a|M)\\rangle` is the halo profile as a
        function of scale, scale factor and halo mass.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
            k (:obj:`float` or `array`): comoving wavenumber.
            a (:obj:`float`): scale factor.
            prof (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`):
                halo profile.

        Returns:
            (:obj:`float` or `array`): integral values evaluated at each
            value of ``k``.
        """
        self._check_mass_def(prof)
        self._get_ingredients(cosmo, a, get_bf=False)
        uk = prof.fourier(cosmo, k, self._mass, a).T
        return self._integrate_over_mf(uk)

    def I_1_1(self, cosmo, k, a, prof):
        """ Solves the integral:

        .. math::
            I^1_1(k,a|u) = \\int dM\\,n(M,a)\\,b(M,a)\\,
            \\langle u(k,a|M)\\rangle,

        where :math:`n(M,a)` is the halo mass function,
        :math:`b(M,a)` is the halo bias, and
        :math:`\\langle u(k,a|M)\\rangle` is the halo profile as a
        function of scale, scale factor and halo mass.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
            k (:obj:`float` or `array`): comoving wavenumber.
            a (:obj:`float`): scale factor.
            prof (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`):
                halo profile.

        Returns:
            (:obj:`float` or `array`): integral values evaluated at each
            value of ``k``.
        """
        self._check_mass_def(prof)
        self._get_ingredients(cosmo, a, get_bf=True)
        uk = prof.fourier(cosmo, k, self._mass, a).T
        return self._integrate_over_mbf(uk)

    def I_0_2(self, cosmo, k, a, prof, *, prof2=None, prof_2pt):
        """ Solves the integral:

        .. math::
            I^0_2(k,a|u,v) = \\int dM\\,n(M,a)\\,
            \\langle u(k,a|M) v(k,a|M)\\rangle,

        where :math:`n(M,a)` is the halo mass function, and
        :math:`\\langle u(k,a|M) v(k,a|M)\\rangle` is the two-point
        moment of the two halo profiles.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
            k (:obj:`float` or `array`): comoving wavenumber.
            a (:obj:`float`): scale factor.
            prof (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`):
                halo profile.
            prof2 (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`): a
                second halo profile. If ``None``, ``prof`` will be used as
                ``prof2``.
            prof_2pt (:class:`~pyccl.halos.profiles_2pt.Profile2pt`):
                a profile covariance object
                returning the the two-point moment of the two profiles
                being correlated.

        Returns:
             (:obj:`float` or `array`): integral values evaluated at each
             value of ``k``.
        """
        if prof2 is None:
            prof2 = prof

        self._check_mass_def(prof, prof2)
        self._get_ingredients(cosmo, a, get_bf=False)
        uk = prof_2pt.fourier_2pt(cosmo, k, self._mass, a, prof, prof2=prof2).T
        return self._integrate_over_mf(uk)

    def I_1_2(self, cosmo, k, a, prof, *, prof2=None, prof_2pt):
        """ Solves the integral:

        .. math::
            I^1_2(k,a|u,v) = \\int dM\\,n(M,a)\\,b(M,a)\\,
            \\langle u(k,a|M) v(k,a|M)\\rangle,

        where :math:`n(M,a)` is the halo mass function,
        :math:`b(M,a)` is the halo bias, and
        :math:`\\langle u(k,a|M) v(k,a|M)\\rangle` is the two-point
        moment of the two halo profiles.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
            k (:obj:`float` or `array`): comoving wavenumber.
            a (:obj:`float`): scale factor.
            prof (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`):
                halo profile.
            prof2 (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`): a
                second halo profile. If ``None``, ``prof`` will be used as
                ``prof2``.
            prof_2pt (:class:`~pyccl.halos.profiles_2pt.Profile2pt`):
                a profile covariance object
                returning the the two-point moment of the two profiles
                being correlated.

        Returns:
             (:obj:`float` or `array`): integral values evaluated at each
             value of ``k``.
        """
        if prof2 is None:
            prof2 = prof

        self._check_mass_def(prof, prof2)
        self._get_ingredients(cosmo, a, get_bf=True)
        uk = prof_2pt.fourier_2pt(cosmo, k, self._mass, a, prof, prof2=prof2).T
        return self._integrate_over_mbf(uk)

    def I_0_22(self, cosmo, k, a, prof, *,
               prof2=None, prof3=None, prof4=None,
               prof12_2pt, prof34_2pt=None):
        """ Solves the integral:

        .. math::
            I^0_{2,2}(k_u,k_v,a|u_{1,2},v_{1,2}) =
            \\int dM\\,n(M,a)\\,
            \\langle u_1(k_u,a|M) u_2(k_u,a|M)\\rangle
            \\langle v_1(k_v,a|M) v_2(k_v,a|M)\\rangle,

        where :math:`n(M,a)` is the halo mass function, and
        :math:`\\langle u(k,a|M) v(k,a|M)\\rangle` is the
        two-point moment of the two halo profiles.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
            k (:obj:`float` or `array`): comoving wavenumber.
            a (:obj:`float`): scale factor.
            prof (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`):
                halo profile.
            prof2 (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`): a
                second halo profile. If ``None``, ``prof`` will be used as
                ``prof2``.
            prof3 (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`): a
                third halo profile. If ``None``, ``prof`` will be used as
                ``prof3``.
            prof4 (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`): a
                fourth halo profile. If ``None``, ``prof2`` will be used as
                ``prof4``.
            prof12_2pt (:class:`~pyccl.halos.profiles_2pt.Profile2pt`):
                a profile covariance object returning the the
                two-point moment of ``prof`` and ``prof2``.
            prof34_2pt (:class:`~pyccl.halos.profiles_2pt.Profile2pt`):
                a profile covariance object returning the the
                two-point moment of ``prof3`` and ``prof4``. If ``None``,
                ``prof12_2pt`` will be used.

        Returns:
             (:obj:`float` or `array`): integral values evaluated at each
             value of ``k``.
        """
        if prof3 is None:
            prof3 = prof
        if prof4 is None:
            prof4 = prof2

        if prof34_2pt is None:
            prof34_2pt = prof12_2pt

        self._check_mass_def(prof, prof2, prof3, prof4)
        self._get_ingredients(cosmo, a, get_bf=False)
        uk12 = prof12_2pt.fourier_2pt(
            cosmo, k, self._mass, a, prof, prof2=prof2).T

        if (prof, prof2, prof12_2pt) == (prof3, prof4, prof34_2pt):
            # 4pt approximation of the same profile
            uk34 = uk12
        else:
            uk34 = prof34_2pt.fourier_2pt(
                cosmo, k, self._mass, a, prof3, prof2=prof4).T

        return self._integrate_over_mf(uk12[None, :, :] * uk34[:, None, :])



"""Test the hashing function of CCL."""
import pytest
import pyccl as ccl
import numpy as np
from collections import OrderedDict

OBJECTS = [ccl.Cosmology,  # class
           (0, 1, 2),      # tuple
           [0, 1, 2],      # list
           set([0, 1, 2]),   # set
           np.arange(3),   # array
           {0: None, 1: None, 2: None},                    # dict
           {0: None, 1: None, 2: {2.1: None, 2.2: None}},  # nested dict
           OrderedDict({0: None, 1: None, 2: None}),       # OrderedDict
           ccl.CosmologyVanillaLCDM(),  # something else
           None,                        # something else
           ]


@pytest.mark.parametrize("obj", OBJECTS)
def test_hashing_smoke(obj):
    assert isinstance(ccl.hash_(obj), int)


def test_hashing_large_array():
    # Hashing ultimately uses the representation of the object.
    # The representation of large numpy arrays only contains the start
    # and the end. We check that the entire array is considered.
    array = np.random.random(64**3).reshape(64, 64, 64)
    array2 = array.copy()
    array2[31, 31, 31] += 1.  # this is now the max value
    vmax = str(array2.max())[:6]
    assert vmax not in repr(array2)  # make sure it doesn't show
    assert ccl.hash_(array) != ccl.hash_(array2)



__all__ = ("MassFuncBocquet16",)

import numpy as np

from . import MassFunc


class MassFuncBocquet16(MassFunc):
    """Implements the mass function of `Bocquet et al. 2016
    <https://arxiv.org/abs/1502.07357>`_. This parametrization accepts
    S.O. masses with :math:`\\Delta = 200` with respect to the matter
    or critical densities, and :math:`\\Delta=500` with respect to the
    critical density.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`):
            a mass definition object, or a name string.
        mass_def_strict (:obj:`bool`): if ``False``, consistency of the mass
            definition will be ignored.
        hydro (:obj:`bool`): if ``False``, use the parametrization found
            using dark-matter-only simulations. Otherwise, include
            baryonic effects (default).
    """
    __repr_attrs__ = __eq_attrs__ = ("mass_def", "mass_def_strict", "hydro",)
    _mass_def_strict_always = True
    name = 'Bocquet16'

    def __init__(self, *,
                 mass_def="200m",
                 mass_def_strict=True,
                 hydro=True):
        self.hydro = hydro
        super().__init__(mass_def=mass_def, mass_def_strict=mass_def_strict)

    def _check_mass_def_strict(self, mass_def):
        return mass_def.name not in ["200m", "200c", "500c"]

    def _setup(self):
        # key: (hydro, mass_def.name)
        vals = {(True, "200m"): (0.228, 2.15, 1.69, 1.30,
                                 0.285, -0.058, -0.366, -0.045),
                (False, "200m"): (0.175, 1.53, 2.55, 1.19,
                                  -0.012, -0.040, -0.194, -0.021),
                (True, "200c"): (0.202, 2.21, 2.00, 1.57,
                                 1.147, 0.375, -1.074, -0.196),
                (False, "200c"): (0.222, 1.71, 2.24, 1.46,
                                  0.269, 0.321, -0.621, -0.153),
                (True, "500c"): (0.180, 2.29, 2.44, 1.97,
                                 1.088, 0.150, -1.008, -0.322),
                (False, "500c"): (0.241, 2.18, 2.35, 2.02,
                                  0.370, 0.251, -0.698, -0.310)}

        key = (self.hydro, self.mass_def.name)
        self.A0, self.a0, self.b0, self.c0, \
            self.Az, self.az, self.bz, self.cz = vals[key]

    def _M200c_M200m(self, cosmo, a):
        # Translates the parameters of M200c to those of M200m
        # for which the base Bocquet16 model is defined.
        z = 1/a - 1
        Omega_m = cosmo.omega_x(a, "matter")
        gamma0 = 3.54E-2 + Omega_m**0.09
        gamma1 = 4.56E-2 + 2.68E-2 / Omega_m
        gamma2 = 0.721 + 3.50E-2 / Omega_m
        gamma3 = 0.628 + 0.164 / Omega_m
        delta0 = -1.67E-2 + 2.18E-2 * Omega_m
        delta1 = 6.52E-3 - 6.86E-3 * Omega_m
        gamma = gamma0 + gamma1 * np.exp(-((gamma2 - z) / gamma3)**2)
        delta = delta0 + delta1 * z
        return gamma, delta

    def _M500c_M200m(self, cosmo, a):
        # Translates the parameters of M500c to those of M200m
        # for which the base Bocquet16 model is defined.
        z = 1/a - 1
        Omega_m = cosmo.omega_x(a, "matter")
        alpha0 = 0.880 + 0.329 * Omega_m
        alpha1 = 1.00 + 4.31E-2 / Omega_m
        alpha2 = -0.365 + 0.254 / Omega_m
        alpha = alpha0 * (alpha1 * z + alpha2) / (z + alpha2)
        beta = -1.7E-2 + 3.74E-3 * Omega_m
        return alpha, beta

    def _get_fsigma(self, cosmo, sigM, a, lnM):
        zp1 = 1./a
        AA = self.A0 * zp1**self.Az
        aa = self.a0 * zp1**self.az
        bb = self.b0 * zp1**self.bz
        cc = self.c0 * zp1**self.cz

        f = AA * ((sigM / bb)**-aa + 1.0) * np.exp(-cc / sigM**2)

        if self.mass_def.name == '200c':
            gamma, delta = self._M200c_M200m(cosmo, a)
            f *= gamma + delta * lnM
        elif self.mass_def.name == '500c':
            alpha, beta = self._M500c_M200m(cosmo, a)
            f *= alpha + beta * lnM

        return f



"""Config file for pytest. Allow caching for faster test completion.
`$ pytest tests/ --use-cache`
"""
from .test_cclobject import all_subclasses, init_decorator
import pyccl


def pytest_addoption(parser):
    parser.addoption("--use-cache", action="store_true", help="Enable cache.")


def pytest_generate_tests(metafunc):
    if metafunc.config.getoption("use_cache"):
        import pyccl
        pyccl.Caching.enable()


# For testing, we want to make sure that all instances of the subclasses
# of `CCLAutoRepr` contain all attributes listed in `__repr_attrs__`.
# We run some things post-init for these subclasses, which are triggered
# during smoke tests.
for sub in list(all_subclasses(pyccl.CCLAutoRepr)):
    sub.__init__ = init_decorator(sub.__init__)
